#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from collections import OrderedDict
from typing import Union, Dict, Callable, Sequence, Any, Optional

import numpy as np
import pandas as pd

__all__ = ['ParameterGenerator']


class ParameterGenerator(object):
    """
    Generator for parameter set generating.

    """

    def __init__(self, seed: Optional[int] = None, **kwargs: Union[Any, Sequence, Callable, Dict]):
        """

        Parameters
        ----------
        seed
            Numpy random seed.
        kwargs
            Parameter candidate.
        """
        if len(kwargs) == 0:
            raise RuntimeError('need parameter candidate')

        np.random.seed(seed)

        self.tuples = OrderedDict()
        self.funcs = OrderedDict()
        self.dicts = OrderedDict()
        self.others = {}

        for k, v in kwargs.items():
            if isinstance(v, (tuple, list, np.ndarray, pd.Series)):
                self.tuples[k] = v
            elif callable(v):
                self.funcs[k] = v
            elif isinstance(v, dict):
                repeat = v['repeat']
                self.dicts[k] = v

                if isinstance(repeat, str):
                    if repeat in self.tuples:
                        self.tuples.move_to_end(repeat, True)
                    if repeat in self.dicts:
                        self.dicts.move_to_end(repeat, True)
                    if repeat in self.funcs:
                        self.funcs.move_to_end(repeat, True)
            else:
                self.others[k] = v

    def __call__(self, num: int, *, factory=None):
        for _ in range(num):
            tmp = {}
            for k, v in self.tuples.items():
                tmp[k] = self._gen(v)

            for k, v in self.funcs.items():
                tmp[k] = v()

            for k, v in reversed(self.dicts.items()):
                data = v['data']
                repeat = v['repeat']
                if 'replace' in v:
                    replace = v['replace']
                else:
                    replace = True

                if isinstance(repeat, (tuple, list, np.ndarray, pd.Series)):
                    repeat = self._gen(repeat)
                elif isinstance(repeat, str):
                    repeat = len(tmp[repeat])

                if isinstance(data, (tuple, list, np.ndarray, pd.Series)):
                    tmp[k] = self._gen(data, repeat, replace)
                elif callable(data):
                    tmp[k] = tuple(data(repeat))

            tmp = dict(self.others, **tmp)
            if factory is not None:
                yield tmp, factory(**tmp)
            else:
                yield tmp

    @staticmethod
    def _gen(item: Sequence, repeat: int = None, replace: bool = True):
        if repeat is not None:
            idx = np.random.choice(len(item), repeat, replace=replace)
            return tuple([item[i] for i in idx])
        else:
            idx = np.random.choice(len(item))
            return item[idx]



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from collections import OrderedDict
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from platform import version as sys_ver
from sys import version as py_ver
from typing import Union, Callable, Any

import numpy as np
import torch

from xenonpy import __version__
from xenonpy.model.training import Trainer, Checker
from xenonpy.model.training.base import BaseExtension

__all__ = ['Persist']


class Persist(BaseExtension):
    """
    Trainer extension for data persistence
    """

    def __init__(self,
                 path: Union[Path, str] = '.',
                 *,
                 model_class: Callable = None,
                 model_params: Union[tuple, dict, any] = None,
                 increment=False,
                 sync_training_step=False,
                 **describe: Any):
        """

        Parameters
        ----------
        path
            Path for model saving.
        model_class
            A factory function for model reconstructing.
            In most case this is the model class inherits from :class:`torch.nn.Module`
        model_params
            The parameters for model reconstructing.
            This can be anything but in general this is a dict which can be used as kwargs parameters.
        increment
            If ``True``, dir name of path will be decorated with a auto increment number,
            e.g. use ``model_dir@1`` for ``model_dir``.
        sync_training_step
            If ``True``, will save ``trainer.training_info`` at each iteration.
            Default is ``False``, only save ``trainer.training_info`` at each epoch.
        describe:
            Any other information to describe this model.
            These information will be saved under model dir by name ``describe.pkl.z``.
        """
        self._model_class: Callable = model_class
        self._model_params: Union[list, dict] = model_params
        self.sync_training_step = sync_training_step
        self.path = path
        self._increment = increment
        self._describe = describe
        self._describe_ = None
        self._checker: Union[Checker, None] = None
        self._tmp_args: list = []
        self._tmp_kwargs: dict = {}
        self._epoch_count = 0

    @property
    def describe(self):
        return self._checker.describe

    @property
    def path(self):
        return str(self._path)

    @path.setter
    def path(self, path: Union[Path, str]):
        if path == '.':
            path = Path(path).resolve()
            self._path = path / path.name
        else:
            self._path = Path(path).resolve()

    @property
    def model_structure(self):
        return self._checker.model_structure

    def get_checkpoint(self, id_: str = None):
        if id_ is not None:
            return self._checker.checkpoints[id_]
        return self._checker.checkpoints.files

    def __call__(self, handle: Any = None, **kwargs: Any):
        if self._checker is None:
            raise RuntimeError('calling of this method only after the model training')
        self._checker(handle=handle, **kwargs)

    def __getitem__(self, item):
        return self._checker[item]

    def on_checkpoint(self, checkpoint: Trainer.checkpoint_tuple, trainer: Trainer) -> None:
        key = checkpoint.id
        value = deepcopy(checkpoint._asdict())
        self._checker.set_checkpoint(**{key: value})

    def step_forward(self, step_info: OrderedDict, trainer: Trainer) -> None:
        if self.sync_training_step:
            training_info = trainer.training_info
            if training_info is not None:
                self._checker(training_info=training_info)
        else:
            epoch = step_info['i_epoch']
            if epoch > self._epoch_count:
                training_info = trainer.training_info
                if training_info is not None:
                    self._epoch_count = epoch
                    self._checker(training_info=training_info)

    def before_proc(self, trainer: Trainer) -> None:
        self._checker = Checker(self._path, increment=self._increment)
        if self._model_class is not None:
            self._checker(model_class=self._model_class)
        if self._model_params is not None:
            self._checker(model_params=self._model_params)
        self._checker.model = trainer.model
        self._describe_ = dict(
            python=py_ver,
            system=sys_ver(),
            numpy=np.__version__,
            torch=torch.__version__,
            xenonpy=__version__,
            device=str(trainer.device),
            start=datetime.now().strftime('%Y/%m/%d %H:%M:%S'),
            finish='N/A',
            time_elapsed='N/A',
            **self._describe,
        )
        self._checker(describe=self._describe_)

    def after_proc(self, trainer: Trainer) -> None:
        self._describe_.update(
            finish=datetime.now().strftime('%Y/%m/%d %H:%M:%S'),
            time_elapsed=str(timedelta(seconds=trainer.timer.elapsed)))
        self._checker.final_state = trainer.model.state_dict()
        self._checker(
            training_info=trainer.training_info,
            describe=self._describe_,
        )



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import math
from typing import Union, Tuple, Callable, Any, Optional

from torch import nn

__all__ = ['LinearLayer', 'SequentialLinear']


class LinearLayer(nn.Module):
    """
    Base NN layer. This is a wrap around PyTorch.
    See here for details: http://pytorch.org/docs/master/nn.html#
    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True, *,
                 dropout: float = 0.,
                 activation_func: Callable = nn.ReLU(),
                 normalizer: Union[float, None] = .1
                 ):
        """
        Parameters
        ----------
        in_features:
            Size of each input sample.
        out_features:
            Size of each output sample
        dropout: float
            Probability of an element to be zeroed. Default: 0.5
        activation_func: func
            Activation function.
        normalizer: func
            Normalization layers
        """
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias)
        self.dropout = nn.Dropout(dropout)
        self.normalizer = None if not normalizer else nn.BatchNorm1d(out_features, normalizer)
        self.activation = None if not activation_func else activation_func

    def forward(self, x):
        _out = self.linear(x)
        if self.dropout:
            _out = self.dropout(_out)
        if self.normalizer:
            _out = self.normalizer(_out)
        if self.activation:
            _out = self.activation(_out)
        return _out


class SequentialLinear(nn.Module):
    """
    Sequential model with linear layers and configurable other hype-parameters.
    e.g. ``dropout``, ``hidden layers``

    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True, *,
                 h_neurons: Union[Tuple[float, ...], Tuple[int, ...]] = (),
                 h_bias: Union[bool, Tuple[bool, ...]] = True,
                 h_dropouts: Union[float, Tuple[float, ...]] = 0.1,
                 h_normalizers: Union[float, None, Tuple[Optional[float], ...]] = 0.1,
                 h_activation_funcs: Union[Callable, None, Tuple[Optional[Callable], ...]] = nn.ReLU(),
                 ):
        """

        Parameters
        ----------
        in_features
            Size of input.
        out_features
            Size of output.
        bias
            Enable ``bias`` in input layer.
        h_neurons
            Number of neurons in hidden layers.
            Can be a tuple of floats. In that case,
            all these numbers will be used to calculate the neuron numbers.
            e.g. (0.5, 0.4, ...) will be expanded as (in_features * 0.5, in_features * 0.4, ...)
        h_bias
            ``bias`` in hidden layers.
        h_dropouts
            Probabilities of dropout in hidden layers.
        h_normalizers
            Momentum of batched normalizers in hidden layers.
        h_activation_funcs
            Activation functions in hidden layers.
        """
        super().__init__()
        self._h_layers = len(h_neurons)
        if self._h_layers > 0:
            if isinstance(h_neurons[0], float):
                tmp = [in_features]
                for i, ratio in enumerate(h_neurons):
                    num = math.ceil(in_features * ratio)
                    tmp.append(num)
                neurons = tuple(tmp)

            elif isinstance(h_neurons[0], int):
                neurons = (in_features,) + tuple(h_neurons)
            else:
                raise RuntimeError('illegal parameter type of <h_neurons>')

            activation_funcs = self._check_input(h_activation_funcs)
            normalizers = self._check_input(h_normalizers)
            dropouts = self._check_input(h_dropouts)
            bias = (bias,) + self._check_input(h_bias)

            for i in range(self._h_layers):
                setattr(self, f'layer_{i}', LinearLayer(
                    in_features=neurons[i],
                    out_features=neurons[i + 1],
                    bias=bias[i],
                    dropout=dropouts[i],
                    activation_func=activation_funcs[i],
                    normalizer=normalizers[i]
                ))

            self.output = nn.Linear(neurons[-1], out_features, bias[-1])
        else:
            self.output = nn.Linear(in_features, out_features, bias)

    def _check_input(self, i):
        if isinstance(i, Tuple):
            if len(i) != self._h_layers:
                raise RuntimeError(
                    f'number of parameter not consistent with number of layers, '
                    f'input is {len(i)} but need to be {self._h_layers}')
            return tuple(i)
        else:
            return tuple([i] * self._h_layers)

    def forward(self, x: Any) -> Any:
        for i in range(self._h_layers):
            x = getattr(self, f'layer_{i}')(x)
        return self.output(x)



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from xenonpy.mdl.base import BaseQuery


class QueryModelsetsWith(BaseQuery):
    queryable = [
        'id',
        'name',
        'describe',
        'deprecated',
        'private',
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query (
                $name_has: [String!]
                $tag_has: [String!]
                $describe_has: [String!]
                $private: Boolean
                $deprecated: Boolean
            ) {{
                queryModelsetsWith(
                    name_has: $name_has
                    tag_has: $tag_has
                    describe_has: $describe_has
                    private: $private
                    deprecated: $deprecated
                ) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class QueryModelsets(BaseQuery):
    queryable = [
        'id',
        'name',
        'describe',
        'deprecated',
        'private',
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($query: [String!]!) {{
                queryModelsets(query: $query) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class GetModelsetDetail(BaseQuery):
    queryable = [
        'id',
        'name',
        'describe',
        'deprecated',
        'private',
        'contributors',
        'owner',
        'sampleCode',
        'tags',
        'count'
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)
        self._return_json = True

    def gql(self, *query_vars: str):
        return f'''
            query ($id: Int!) {{
                getModelsetDetail(id: $id) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class ListModelsets(BaseQuery):
    queryable = [
        'id',
        'name',
        'describe',
        'deprecated',
        'private',
    ]

    def __init__(self, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables={}, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query {{
                listModelsets {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class CreateModelset(BaseQuery):
    queryable = [
        'id',
        'name',
        'describe',
        'deprecated',
        'private',
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)
        self._return_json = True

    def gql(self, *query_vars: str):
        return f'''
            mutation ($with_: CreateModelsetInput!) {{
                createModelset (with_: $with_) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class UpdateModelset(BaseQuery):
    queryable = [
        'id',
        'name',
        'describe',
        'deprecated',
        'private',
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)
        self._return_json = True

    def gql(self, *query_vars: str):
        return f'''
            mutation (
                $id: Int!
                $with_: UpdateModelsetInput!
            ) {{
                updateModelset (
                    id: $id
                    with_: $with_
                ) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''



# IQSPR with focus on molecule variety: bring in new initial molecules from reservoir in every step of SMC and allow mutation based on fragmentation

import numpy as np
import pandas as pd

from rdkit import Chem
from xenonpy.inverse.iqspr import NGram
from xenonpy.inverse.base import BaseSMC, BaseProposal, BaseLogLikelihood

class SMCError(Exception):
    """Base exception for SMC classes"""
    pass

class IQSPR_VF(BaseSMC):

    def __init__(self, *, estimator, modifier):
        """
        SMC iqspr runner.
        Parameters
        ----------
        estimator : BaseLogLikelihood or BaseLogLikelihoodSet
            Log likelihood estimator for given SMILES.
        modifier : BaseProposal
            Modify given SMILES to new ones.
        """
        self._proposal = modifier
        self._log_likelihood = estimator

    def resample(self, sims, size, p):
        return np.random.choice(sims, size=size, p=p)

    @property
    def modifier(self):
        return self._proposal

    @modifier.setter
    def modifier(self, value):
        self._proposal = value

    @property
    def estimator(self):
        return self._log_likelihood

    @estimator.setter
    def estimator(self, value):
        self._log_likelihood = value
        
    def fragmenting(self, smis):
        frag_list = []
        
        for smi in smis:
            decomp = Chem.Recap.RecapDecompose(Chem.MolFromSmiles(smi))
            first_gen = [node.mol for node in decomp.children.values()]
            frag_list += [Chem.MolToSmiles(x) for x in first_gen]
                
        return list(set(frag_list))
    
    def combine_fragments(self, smis_base, smis_frag):
        """
        combine two SMILES strings with '*' as connection points
        Parameters
        ----------
        smis_base: str
            SMILES for combining.
            If no '*', assume connection point at the end.
            If more than one '*', the first will be picked if it's not the 1st character.
        smis_frag: str
            SMILES for combining.
            If no '*', assume connection point at the front.
            If more than one '*', the first will be picked.
        """

        # prepare NGram object for use of ext. SMILES
        ngram = NGram()

        # check position of '*'
        mols_base = Chem.MolFromSmiles(smis_base)
        if mols_base is None:
            raise RuntimeError('Invalid base SMILES!')
        idx_base = [i for i in range(mols_base.GetNumAtoms()) if mols_base.GetAtomWithIdx(i).GetSymbol() == '*']

        # rearrange base SMILES to avoid 1st char = '*'
        if len(idx_base) == 1 and idx_base[0] == 0:
            smis_base_head = Chem.MolToSmiles(mols_base,rootedAtAtom=1)
        elif len(idx_base) == 0:
            smis_base_head = smis_base + '*'
        else:
            smis_base_head = smis_base

        # converge base to ext. SMILES and pick insertion location
        esmi_base = ngram.smi2esmi(smis_base_head)
        esmi_base = esmi_base[:-1]
        idx_base = esmi_base.index[esmi_base['esmi'] == '*'].tolist()
        if idx_base[0] == 0:
            idx_base = idx_base[1]
        else:
            idx_base = idx_base[0]

        # rearrange fragment to have 1st char = '*' and convert to ext. SMILES
        mols_frag = Chem.MolFromSmiles(smis_frag)
        if mols_frag is None:
            raise RuntimeError('Invalid frag SMILES!')
        idx_frag = [i for i in range(mols_frag.GetNumAtoms()) if mols_frag.GetAtomWithIdx(i).GetSymbol() == '*']
        if len(idx_frag) == 0:
            esmi_frag = ngram.smi2esmi(smis_frag)
            # remove last '!'
            esmi_frag = esmi_frag[:-1]
        else:
            esmi_frag = ngram.smi2esmi(Chem.MolToSmiles(mols_frag,rootedAtAtom=idx_frag[0]))
            # remove leading '*' and last '!'
            esmi_frag = esmi_frag[1:-1]

        # check open rings of base SMILES
        nRing_base = esmi_base['n_ring'].loc[idx_base]

        # re-number rings in fragment SMILES
        esmi_frag['n_ring'] = esmi_frag['n_ring'] + nRing_base

        # delete '*' at the insertion location
        esmi_base = esmi_base.drop(idx_base).reset_index(drop=True)

        # combine base with the fragment
        ext_smi = pd.concat([esmi_base.iloc[:idx_base], esmi_frag, esmi_base.iloc[idx_base:]]).reset_index(drop=True)
        new_pd_row = {'esmi': '!', 'n_br': 0, 'n_ring': 0, 'substr': ['!']}
        ext_smi.append(new_pd_row, ignore_index=True)

        return ngram.esmi2smi(ext_smi)

    def __call__(self, reservoir, beta, size=100, *, samples=None, ratio=0.5, p_frag=0.1, yield_lpf=False):
        """
        Run SMC
        Parameters
        ----------
        reservoir: list of object
            Samples to be drawn as new initial molecules in each step of SMC
        beta: list/1D-numpy of float or pd.Dataframe
            Annealing parameters for each step.
            If pd.Dataframe, column names should follow keys of mdl in BaseLogLikeihood or BaseLogLikelihoodSet
        size: int
            Sample size for each draw.
        samples: list of object
            Initial samples.
        ratio: float
            Ratio of molecules to be replaced from reservoir in each step of SMC
        p_frag: float
            Probability of performing a fragment based mutation
        yield_lpf : bool
            Yield estimated log likelihood, probability and frequency of each samples. Default is ``False``.
        Yields
        -------
        samples: list of object
            New samples in each SMC iteration.
        llh: np.ndarray float
            Estimated values of log-likelihood of each samples.
            Only yield when ``yield_lpf=Ture``.
        p: np.ndarray of float
            Estimated probabilities of each samples.
            Only yield when ``yield_lpf=Ture``.
        freq: np.ndarray of float
            The number of unique samples in original samples.
            Only yield when ``yield_lpf=Ture``.
        """

        # initial samples will be randomly picked from resevoir if samples are provided
        if samples is None:
            samples = np.random.choice(reservoir,size=size).tolist()
        # refill samples if len(samples) not equals given size
        elif len(samples) < size:
            samples = samples + np.random.choice(reservoir,size=size-len(samples)).tolist()
            
        res_size = int(size*ratio)
        smc_size = size - res_size

        try:
            unique, frequency = self.unique(samples)
            ll = self.log_likelihood(unique)

            if isinstance(beta, pd.DataFrame):
                beta = beta[ll.columns].values
            else:
                # assume only one row for beta (1-D list or numpy vector)
                beta = np.transpose(np.repeat([beta], ll.shape[1], axis=0))

            w = np.dot(ll.values, beta[0]) + np.log(frequency)
            w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
            p = np.exp(w - w_sum)
            if yield_lpf:
                yield unique, ll, p, frequency
            else:
                yield unique

            # beta = np.delete(beta,0,0) #remove first "row"

        except SMCError as e:
            self.on_errors(0, samples, e)
        except Exception as e:
            raise e

        # translate between input representation and execute environment representation.
        # make sure beta is not changed (np.delete will return the deleted version, without changing original vector)
        for i, step in enumerate(np.delete(beta, 0, 0)):
            try:
                re_samples = self.resample(unique, smc_size, p)
                if np.random.random() < p_frag:
                    frag_list = self.fragmenting(re_samples + np.random.choice(reservoir,size=res_size).tolist())
                    samples = [self.combine_fragments(np.random.choice(frag_list), np.random.choice(frag_list)) for _ in range(size)]
                else:
                    samples = self.proposal(re_samples + np.random.choice(reservoir,size=res_size).tolist()) 

                unique, frequency = self.unique(samples)

                # annealed likelihood in log - adjust with copy counts
                ll = self.log_likelihood(unique)
                w = np.dot(ll.values, step) + np.log(frequency)
                w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
                p = np.exp(w - w_sum)
                if yield_lpf:
                    yield unique, ll, p, frequency
                else:
                    yield unique

            except SMCError as e:
                self.on_errors(i + 1, samples, e)
            except Exception as e:
                raise e
                



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.


import os
import tarfile
from pathlib import Path
from typing import List, Union

import pandas as pd
import requests
from sklearn.base import BaseEstimator
from tqdm import tqdm

from xenonpy.utils import TimedMetaClass
from .base import BaseQuery
from .descriptor import QueryDescriptors, QueryDescriptorsWith, UpdateDescriptor, CreateDescriptor, ListDescriptors, \
    GetDescriptorDetail
from .method import QueryMethods, QueryMethodsWith, UpdateMethod, CreateMethod, ListMethods, GetMethodDetail
from .model import QueryModelDetails, QueryModelDetailsWith, UploadModel, GetTrainingInfo, GetTrainingEnv, \
    GetSupplementary, GetModelUrls, GetModelUrl, GetModelDetails, GetModelDetail, ListModelsWithProperty, \
    ListModelsWithModelset, ListModelsWithMethod, ListModelsWithDescriptor
from .modelset import QueryModelsets, QueryModelsetsWith, UpdateModelset, CreateModelset, ListModelsets, \
    GetModelsetDetail
from .property import QueryPropertiesWith, QueryProperties, UpdateProperty, CreateProperty, ListProperties, \
    GetPropertyDetail

__all__ = ['MDL', 'GetVersion', 'QueryModelsetsWith', 'QueryModelsets', 'QueryModelDetailsWith', 'QueryModelDetails',
           'UpdateModelset', 'UploadModel', 'GetModelsetDetail', 'GetModelDetail', 'GetModelDetails', 'GetModelUrls',
           'GetModelUrl', 'GetTrainingInfo', 'GetSupplementary', 'GetTrainingEnv', 'ListModelsets',
           'ListModelsWithDescriptor', 'ListModelsWithMethod', 'ListModelsWithModelset', 'ListModelsWithProperty',
           'QueryPropertiesWith', 'QueryProperties', 'GetPropertyDetail', 'ListProperties', 'CreateProperty',
           'CreateModelset', 'UpdateProperty', 'QueryDescriptorsWith', 'QueryDescriptors', 'QueryMethodsWith',
           'QueryMethods', 'UpdateDescriptor', 'UpdateMethod', 'ListDescriptors', 'ListMethods', 'GetMethodDetail',
           'GetDescriptorDetail', 'CreateDescriptor', 'CreateMethod']


class GetVersion(BaseQuery):
    queryable = []

    def __init__(self, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables={}, api_key=api_key, endpoint=endpoint)
        self._return_json = True

    def gql(self, *query_vars: str):
        return '''
            query {
                getVersion 
            }
            '''


class MDL(BaseEstimator, metaclass=TimedMetaClass):
    def __init__(self, *, api_key: str = 'anonymous.user.key', endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Access key.
        endpoint:
            Url to XenonPy.MDL api server.
        """
        self._endpoint = endpoint
        self._api_key = api_key

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, e):
        self._endpoint = e

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, k):
        """"""
        self._api_key = k

    @property
    def version(self):
        return GetVersion()()

    def __call__(self, *query: str,
                 modelset_has: Union[List[str]] = None,
                 property_has: Union[List[str]] = None,
                 descriptor_has: Union[List[str]] = None,
                 method_has: Union[List[str]] = None,
                 lang_has: Union[List[str]] = None,
                 regression: bool = None,
                 transferred: bool = None,
                 deprecated: bool = None,
                 succeed: bool = None,
                 ) -> Union[QueryModelDetails, QueryModelDetailsWith]:
        """
        Query models with specific keywords.

        Parameters
        ----------
        query
            Lowercase string for database querying.
            This is a fuzzy searching, any information contains given string will hit.
        modelset_has
            The part of a model set's name.
            For example, ``modelset_has='test`` will hit ``*test*``
        property_has
            A part of the name of property.
        descriptor_has
            A part of the name of descriptor.
        method_has
            A part of the name of training method.
        lang_has
            A part of the name of programming language.
        regression
            If``True``, searching in regression models,
            else, searching in classification models.
            Default is ``True``.
        deprecated
            Model with this mark is deprecated.
        transferred
            If ``True``, searching in transferred models.
            Default is ``False``.
        succeed
            If ``True``, searching in succeed models.
            Default is ``True``.

        Returns
        -------
        query
            Querying object.
        """

        if len(query) > 0:
            return QueryModelDetails(dict(query=query), api_key=self.api_key, endpoint=self.endpoint)
        else:
            variables = dict(
                modelset_has=modelset_has,
                property_has=property_has,
                descriptor_has=descriptor_has,
                method_has=method_has,
                lang_has=lang_has,
                regression=regression,
                transferred=transferred,
                deprecated=deprecated,
                succeed=succeed
            )
            variables = {k: v for k, v in variables.items() if v is not None}

            return QueryModelDetailsWith(variables, api_key=self.api_key, endpoint=self.endpoint)

    def upload_model(self, *,
                     modelset_id: int,
                     describe: dict,
                     training_env: dict = None,
                     training_info: dict = None,
                     supplementary: dict = None) -> UploadModel:
        """
        Upload model to XenonPy.MDL server.

        Parameters
        ----------
        modelset_id
        describe
        training_env
        training_info
        supplementary

        Returns
        -------
        query
            Querying object.

        """
        variables = dict(
            model=None,
            id=modelset_id,
            describe=describe,
            training_env=training_env,
            training_info=training_info,
            supplementary=supplementary,
        )
        variables = {k: v for k, v in variables.items() if v is not None}

        return UploadModel(variables, api_key=self.api_key, endpoint=self.endpoint)

    def get_training_info(self, model_id: int) -> GetTrainingInfo:
        """
        Get training information, e.g. ``train_loss``.

        Parameters
        ----------
        model_id
            Model id.

        Returns
        -------
        query
            Querying object.
        """

        return GetTrainingInfo({'id': model_id}, api_key=self.api_key, endpoint=self.endpoint)

    def get_training_env(self, model_id: int) -> GetTrainingEnv:
        """
        Get model's training environments.

        Parameters
        ----------
        model_id
            Model id.

        Returns
        -------
        query
            Querying object.
        """
        return GetTrainingEnv({'id': model_id}, api_key=self.api_key, endpoint=self.endpoint)

    def get_supplementary(self, *, model_id: int) -> GetSupplementary:
        """
        Get extract information about the model.
        For regression model this should be *prediction vs observation* data.
        For classification model, this will be the *confusion matrix*.

        Parameters
        ----------
        model_id
            Model id.

        Returns
        -------
        query
            Querying object.
        """
        return GetSupplementary({'id': model_id}, api_key=self.api_key, endpoint=self.endpoint)

    def get_model_urls(self, *model_ids: int) -> Union[GetModelUrl, GetModelUrls]:
        """
        Get download url by model id.

        Parameters
        ----------
        model_ids
            Model id.

        Returns
        -------
        query
            Querying object.
        """
        if len(model_ids) == 0:
            raise RuntimeError('input is not non-able')
        if len(model_ids) == 1:
            return GetModelUrl({'id': model_ids[0]}, api_key=self.api_key, endpoint=self.endpoint)
        else:
            return GetModelUrls({'ids': model_ids}, api_key=self.api_key, endpoint=self.endpoint)

    def get_model_detail(self, model_id: int) -> GetModelDetail:
        """
        Get model detail by model id.

        Parameters
        ----------
        model_id
            Model id.

        Returns
        -------
        query
            Querying object.
        """
        return GetModelDetail({'id': model_id}, api_key=self.api_key, endpoint=self.endpoint)

    def get_model_details(self, model_ids: List[int]) -> GetModelDetails:
        """
        Get multiply model detail by their ids in one querying.

        Parameters
        ----------
        model_ids
            Model id.

        Returns
        -------
        query
            Querying object.
        """
        return GetModelDetails({'ids': model_ids}, api_key=self.api_key, endpoint=self.endpoint)

    def list_models_with_property(self, name: str) -> ListModelsWithProperty:
        """
        List all models relevant with given property.

        Parameters
        ----------
        name
            Property name

        Returns
        -------
        query
            Querying object.
        """
        return ListModelsWithProperty({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def list_models_with_modelset(self, name: str) -> ListModelsWithModelset:
        """
        List all models relevant with given modelset.

        Parameters
        ----------
        name
            Modelset name

        Returns
        -------
        query
            Querying object.
        """
        return ListModelsWithModelset({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def list_models_with_method(self, name: str) -> ListModelsWithMethod:
        """
        List all models relevant with given method.

        Parameters
        ----------
        name
            Method name

        Returns
        -------
        query
            Querying object.
        """
        return ListModelsWithMethod({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def list_models_with_descriptor(self, name: str) -> ListModelsWithDescriptor:
        """
        List all models relevant with given descriptor.

        Parameters
        ----------
        name
            Descriptor name

        Returns
        -------
        query
            Querying object.
        """
        return ListModelsWithDescriptor({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def query_modelsets(self, query: str = None, *,
                        name_has: Union[List[str]] = None,
                        tag_has: Union[List[str]] = None,
                        describe_has: Union[List[str]] = None,
                        private: bool = None,
                        deprecated: bool = None,
                        ) -> Union[QueryModelsets, QueryModelsetsWith]:
        """
        Query modelsets with specific keywords.

        Parameters
        ----------
        query
            Lowercase string for database querying.
            This is a fuzzy searching, any information contains given string will hit.
        name_has
            The part of a model set's name.
            For example, ``modelset_has='test`` will hit ``*test*``
        tag_has
            A part of the name of property.
        describe_has
            A part of the name of descriptor.
        private
            If``True``, searching in regression models,
            else, searching in classification models.
            Default is ``True``.
        deprecated
            Model with this mark is deprecated.
        deprecated
            If ``True``, searching in transferred models.
            Default is ``False``.

        Returns
        -------
        query
            Querying object.
        """

        if query is not None:
            return QueryModelsets(dict(query=query), api_key=self.api_key, endpoint=self.endpoint)
        else:
            variables = dict(
                name_has=name_has,
                tag_has=tag_has,
                describe_has=describe_has,
                private=private,
                deprecated=deprecated,
            )
            variables = {k: v for k, v in variables.items() if v is not None}

            return QueryModelsetsWith(variables, api_key=self.api_key, endpoint=self.endpoint)

    def update_modelset(self, *,
                        modelset_id: int,
                        name: str = None,
                        describe: str = None,
                        sample_code: str = None,
                        tags: List[str] = None,
                        private: bool = None,
                        deprecated: bool = None
                        ) -> UpdateModelset:
        """
        Update modelset information by modelset id.

        Parameters
        ----------
        modelset_id
        name
        describe
        sample_code
        tags
        private
        deprecated

        Returns
        -------
        query
            Querying object.
        """
        with_ = dict(
            name=name,
            describe=describe,
            sample_code=sample_code,
            tags=tags,
            private=private,
            deprecated=deprecated
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return UpdateModelset({'id': modelset_id, 'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def create_modelset(self, *,
                        name: str,
                        describe: str = None,
                        sample_code: str = None,
                        tags: List[str] = None,
                        private: bool = False
                        ):
        """
        Create modelset.

        Parameters
        ----------
        name
        describe
        sample_code
        tags
        private

        Returns
        -------
        query
            Querying object.
        """
        with_ = dict(
            name=name,
            describe=describe,
            sample_code=sample_code,
            tags=tags,
            private=private
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return CreateModelset({'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def list_modelsets(self) -> ListModelsets:
        """
        List all modelsets.

        Returns
        -------
        query
            Querying object.
        """
        return ListModelsets(api_key=self.api_key, endpoint=self.endpoint)

    def get_modelset_detail(self, modelset_id: int) -> GetModelsetDetail:
        """
        Get modelset detail by modelset id.

        Parameters
        ----------
        modelset_id

        Returns
        -------
        query
            Querying object.
        """
        return GetModelsetDetail({'id': modelset_id}, api_key=self.api_key, endpoint=self.endpoint)

    def query_descriptors(self, query: str = None, *,
                          name_has: Union[List[str]] = None,
                          fullname_has: Union[List[str]] = None,
                          describe_has: Union[List[str]] = None,
                          ) -> Union[QueryDescriptors, QueryDescriptorsWith]:
        """
        Query descriptors with specific keywords.

        Parameters
        ----------
        query
            Lowercase string for database querying.
            This is a fuzzy searching, any information contains given string will hit.
        name_has
            The part of a model set's name.
            For example, ``modelset_has='test`` will hit ``*test*``
        fullname_has
            Part of the full name.
        describe_has
            Part of the descriptor.

        Returns
        -------
        query
            Querying object.
        """

        if query is not None:
            return QueryDescriptors(dict(query=query), api_key=self.api_key, endpoint=self.endpoint)
        else:
            variables = dict(
                name_has=name_has,
                fullName_has=fullname_has,
                describe_has=describe_has,
            )
            variables = {k: v for k, v in variables.items() if v is not None}

            return QueryDescriptorsWith(variables, api_key=self.api_key, endpoint=self.endpoint)

    def update_descriptor(self, *,
                          name: str,
                          new_name: str = None,
                          describe: str = None,
                          fullname: str = None,
                          ) -> UpdateDescriptor:
        """
        Update descriptor information by name.

        Parameters
        ----------
        name
        new_name
        describe
        fullname

        Returns
        -------
        query
            Querying object.
        """

        with_ = dict(
            name=new_name,
            describe=describe,
            fullName=fullname,
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return UpdateDescriptor({'name': name, 'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def create_descriptor(self, *,
                          name: str,
                          describe: str = None,
                          fullname: str = None,
                          ) -> CreateDescriptor:
        """
        Create descriptor.

        Parameters
        ----------
        name
        describe
        fullname

        Returns
        -------
        query
            Querying object.
        """

        with_ = dict(
            name=name,
            describe=describe,
            fullName=fullname,
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return CreateDescriptor({'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def list_descriptors(self) -> ListDescriptors:
        """
        List all descriptor.

        Returns
        -------
        query
            Querying object.
        """
        return ListDescriptors(api_key=self.api_key, endpoint=self.endpoint)

    def get_descriptor_detail(self, name: str) -> GetDescriptorDetail:
        """
        Get descriptor detail by name.

        Parameters
        ----------
        name

        Returns
        -------
        query
            Querying object.
        """
        return GetDescriptorDetail({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def query_methods(self, query: str = None, *,
                      name_has: Union[List[str]] = None,
                      fullname_has: Union[List[str]] = None,
                      describe_has: Union[List[str]] = None,
                      ) -> Union[QueryMethods, QueryMethodsWith]:
        """
        Query methods with specific keywords.

        Parameters
        ----------
        query
            Lowercase string for database querying.
            This is a fuzzy searching, any information contains given string will hit.
        name_has
            The part of a model set's name.
            For example, ``modelset_has='test`` will hit ``*test*``
        fullname_has
            Part of the full name.
        describe_has
            Part of the descriptor.

        Returns
        -------
        query
            Querying object.
        """

        if query is not None:
            return QueryMethods(dict(query=query), api_key=self.api_key, endpoint=self.endpoint)
        else:
            variables = dict(
                name_has=name_has,
                fullName_has=fullname_has,
                describe_has=describe_has,
            )
            variables = {k: v for k, v in variables.items() if v is not None}

            return QueryMethodsWith(variables, api_key=self.api_key, endpoint=self.endpoint)

    def update_method(self, *,
                      name: str,
                      new_name: str = None,
                      describe: str = None,
                      fullname: str = None,
                      ) -> UpdateMethod:
        """
        Update method information by name.

        Parameters
        ----------
        name
        new_name
        describe
        fullname

        Returns
        -------
        query
            Querying object.
        """

        with_ = dict(
            name=new_name,
            describe=describe,
            fullName=fullname,
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return UpdateMethod({'name': name, 'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def create_method(self, *,
                      name: str,
                      describe: str = None,
                      fullname: str = None,
                      ) -> CreateMethod:
        """
        Create method.

        Parameters
        ----------
        name
        describe
        fullname

        Returns
        -------
        query
            Querying object.
        """

        with_ = dict(
            name=name,
            describe=describe,
            fullName=fullname,
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return CreateMethod({'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def list_methods(self) -> ListMethods:
        """
        List all methods.

        Returns
        -------
        query
            Querying object.
        """
        return ListMethods(api_key=self.api_key, endpoint=self.endpoint)

    def get_method_detail(self, name: str) -> GetMethodDetail:
        """
        Get method detail by name.

        Parameters
        ----------
        name

        Returns
        -------
        query
            Querying object.
        """
        return GetMethodDetail({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def query_properties(self, query: str = None, *,
                         name_has: Union[List[str]] = None,
                         fullname_has: Union[List[str]] = None,
                         describe_has: Union[List[str]] = None,
                         symbol_has: Union[List[str]] = None,
                         unit_has: Union[List[str]] = None,
                         ) -> Union[QueryProperties, QueryPropertiesWith]:
        """
        Query properties with specific keywords.

        Parameters
        ----------
        query
            Lowercase string for database querying.
            This is a fuzzy searching, any information contains given string will hit.
        name_has
            The part of a model set's name.
            For example, ``modelset_has='test`` will hit ``*test*``
        fullname_has
            Part of the full name.
        describe_has
            Part of the descriptor.
        symbol_has
            Part of the symbol.
        unit_has
            Part of the unit.

        Returns
        -------
        query
            Querying object.
        """

        if query is not None:
            return QueryProperties(dict(query=query), api_key=self.api_key, endpoint=self.endpoint)
        else:
            variables = dict(
                name_has=name_has,
                fullName_has=fullname_has,
                describe_has=describe_has,
                symbol_has=symbol_has,
                unit_has=unit_has,
            )
            variables = {k: v for k, v in variables.items() if v is not None}

            return QueryPropertiesWith(variables, api_key=self.api_key, endpoint=self.endpoint)

    def update_property(self, *,
                        name: str,
                        new_name: str = None,
                        describe: str = None,
                        fullname: str = None,
                        symbol: str = None,
                        unit: str = None,
                        ) -> UpdateProperty:
        """
        Update property information by name.

        Parameters
        ----------
        name
        new_name
        describe
        fullname
        symbol
        unit

        Returns
        -------
        query
            Querying object.
        """
        with_ = dict(
            name=new_name,
            describe=describe,
            fullName=fullname,
            symbol=symbol,
            priunitvate=unit,
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return UpdateProperty({'name': name, 'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def create_property(self, *,
                        name: str,
                        describe: str = None,
                        fullname: str = None,
                        symbol: str = None,
                        unit: str = None,
                        ) -> CreateProperty:
        """
        Create property.

        Parameters
        ----------
        name
        describe
        fullname
        symbol
        unit

        Returns
        -------
        query
            Querying object.
        """

        with_ = dict(
            name=name,
            describe=describe,
            fullName=fullname,
            symbol=symbol,
            unit=unit
        )
        with_ = {k: v for k, v in with_.items() if v is not None}

        return CreateProperty({'with_': with_}, api_key=self.api_key, endpoint=self.endpoint)

    def list_properties(self) -> ListProperties:
        """
        List all properties.

        Returns
        -------
        query
            Querying object.
        """
        return ListProperties(api_key=self.api_key, endpoint=self.endpoint)

    def get_property_detail(self, name: str):
        """
        Get property detail by name.

        Parameters
        ----------
        name

        Returns
        -------
        query
            Querying object.
        """
        return GetPropertyDetail({'name': name}, api_key=self.api_key, endpoint=self.endpoint)

    def pull(self, *model_ids: Union[int, pd.Series, pd.DataFrame],
             save_to: str = '.') -> pd.DataFrame:
        """
        Download model(s) from XenonPy.MDL server.

        Parameters
        ----------
        model_ids
            Model ids.
            It can be given by a dataframe.
            In this case, the column with name ``id`` will be used.
        save_to
            Path to save models.

        Returns
        -------

        """
        if len(model_ids) == 0:
            raise RuntimeError('input can not be empty')
        if len(model_ids) == 1:
            if isinstance(model_ids[0], pd.Series):
                model_ids = model_ids[0].tolist()

            if isinstance(model_ids[0], pd.DataFrame):
                model_ids = model_ids[0]['id'].tolist()

        ret = self.get_model_urls(*model_ids)('id', 'url')
        urls = ret['url'].tolist()

        path_list = []
        for url in tqdm(urls):
            path = '/'.join(url.split('/')[4:])
            file = Path(save_to).resolve() / path
            file.parent.mkdir(parents=True, exist_ok=True)
            file = str(file)
            r = requests.get(url=url)
            with open(file, 'wb') as f:
                f.write(r.content)
            path = file[:-7]
            tarfile.open(file).extractall(path=path)
            os.remove(file)
            path_list.append(path)

        ret['model'] = path_list
        return ret.drop(columns='url')



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from collections import OrderedDict, namedtuple
from copy import deepcopy
from pathlib import Path
from typing import Union, Tuple, List, Any, Dict, Callable

import numpy as np
import pandas as pd
import torch
from torch.nn import Module
from torch.optim.lr_scheduler import _LRScheduler, ReduceLROnPlateau
from torch.utils.data import DataLoader
from tqdm import tqdm

from xenonpy.model.training import ClipValue, ClipNorm, Checker
from xenonpy.model.training.base import BaseOptimizer, BaseLRScheduler, BaseRunner
from xenonpy.utils import camel_to_snake

__all__ = ['Trainer']


class Trainer(BaseRunner):
    checkpoint_tuple = namedtuple('checkpoint', 'id iterations model_state')
    results_tuple = namedtuple('results', 'total_epochs device training_info checkpoints model')

    def __init__(self,
                 *,
                 loss_func: torch.nn.Module = None,
                 optimizer: BaseOptimizer = None,
                 model: Module = None,
                 lr_scheduler: BaseLRScheduler = None,
                 clip_grad: Union[ClipNorm, ClipValue] = None,
                 epochs: int = 200,
                 cuda: Union[bool, str, torch.device] = False,
                 non_blocking: bool = False,
                 ):
        """
        NN model trainer.

        Parameters
        ----------
        loss_func
            Loss function.
        optimizer
            Optimizer for model parameters tuning.
        model
            Pytorch NN model.
        lr_scheduler
            Learning rate scheduler.
        clip_grad
            Clip grad before each optimize.
        epochs
            Number of iterations.
        cuda
            Set training device(s).
        non_blocking
            When non_blocking is ``True``,
            it tries to convert/move asynchronously with respect to the host if possible.
        """
        super().__init__(cuda=cuda)
        # None able
        self._clip_grad = clip_grad
        self.epochs = epochs
        self.non_blocking = non_blocking

        # loss function placeholder
        self._loss_func = None
        self._loss_type = None

        # model placeholder
        self._model = None
        self._init_states = None

        # optimizer placeholder
        self._optim = None
        self._optimizer = None
        self._optimizer_state = None

        # lr_scheduler placeholder
        self._scheduler = None
        self._lr_scheduler = None

        # init private vars
        self._early_stopping: Tuple[bool, str] = (False, '')
        self._checkpoints: Dict[Union[int, str], Trainer.checkpoint_tuple] = OrderedDict()
        self._training_info: List[OrderedDict] = []
        self._total_its: int = 0  # of total iterations
        self._total_epochs: int = 0  # of total epochs

        self._x_val = None
        self._y_val = None
        self._validate_dataset = None

        # init
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.loss_func = loss_func

    @property
    def loss_type(self):
        return self._loss_type

    @property
    def total_epochs(self):
        return self._total_epochs

    @property
    def total_iterations(self):
        return self._total_its

    @property
    def x_val(self):
        return self._x_val

    @property
    def y_val(self):
        return self._y_val

    @property
    def validate_dataset(self):
        return self._validate_dataset

    @property
    def loss_func(self):
        return self._loss_func

    @loss_func.setter
    def loss_func(self, loss_func):
        if loss_func is not None:
            self._loss_func = loss_func
            self._loss_type = 'train_' + camel_to_snake(loss_func.__class__.__name__)

    @property
    def training_info(self):
        if len(self._training_info) > 0:
            return pd.DataFrame(data=self._training_info)
        return None

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, v):
        self._device = self.check_device(v)
        self.model = None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        if model is not None:
            if isinstance(model, torch.nn.Module):
                self.reset(to=model)
            else:
                raise TypeError(
                    'parameter `m` must be a instance of <torch.nn.modules> but got %s' % type(model))

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer):
        if optimizer is not None:
            self._optim = optimizer

        if self._optim is not None and self._model is not None:
            self._optimizer = self._optim(self._model.parameters())
            self._optimizer_state = deepcopy(self._optimizer.state_dict())

            self.lr_scheduler = None

    @property
    def lr_scheduler(self):
        return self._lr_scheduler

    @lr_scheduler.setter
    def lr_scheduler(self, scheduler):
        if scheduler is not None:
            self._scheduler = scheduler

        if self._scheduler is not None and self._optimizer is not None:
            self._lr_scheduler: Union[_LRScheduler, None] = self._scheduler(self._optimizer)

    @property
    def clip_grad(self):
        return self._clip_grad

    @clip_grad.setter
    def clip_grad(self, fn):
        self._clip_grad = fn

    @property
    def checkpoints(self):
        return self._checkpoints

    def get_checkpoint(self, checkpoint: Union[int, str] = None):
        if checkpoint is None:
            return list(self._checkpoints.keys())
        if isinstance(checkpoint, int):
            id_ = f'cp_{checkpoint}'
            return self._checkpoints[id_]
        if isinstance(checkpoint, str):
            return self._checkpoints[checkpoint]
        raise TypeError(f'parameter <cp> must be str or int but got {checkpoint.__class__}')

    def set_checkpoint(self, id_: str = None):
        if id_ is None:
            id_ = f'cp_{self._total_its}'
        cp = self.checkpoint_tuple(
            id=id_,
            iterations=self._total_its,
            model_state=deepcopy(self._model.state_dict()),
        )

        self._checkpoints[id_] = cp
        self._on_checkpoint(checkpoint=cp, trainer=self, training=True)

    def early_stop(self, msg: str):
        self._early_stopping = (True, msg)

    def reset(self, *, to: Union[Module, int, str] = None, remove_checkpoints: bool = True):
        """
        Reset trainer.
        This will reset all trainer states and drop all training step information.

        Parameters
        ----------
        to: Union[bool, Module]
            Bind trainer to the given model or reset current model to it's initialization states.
        remove_checkpoints
            Set to ``True`` to remove all checkpoints when resetting. Default ``true``.
        """
        self._training_info = []
        self._total_its = 0
        self._total_epochs = 0
        self._early_stopping = (False, '')

        if isinstance(to, Module):
            self._model = to.to(self._device, non_blocking=self.non_blocking)
            self._init_states = deepcopy(to.state_dict())

            self.optimizer = None
            self.lr_scheduler = None
        elif isinstance(to, (int, str)):
            cp = self.get_checkpoint(to)
            self._model.load_state_dict(cp.model_state)
        elif to is None:
            self._model.load_state_dict(self._init_states)
            self._optimizer.load_state_dict(self._optimizer_state)
        else:
            raise TypeError(f'parameter <to> must be torch.nnModule, int, or str but got {type(to)}')

        if remove_checkpoints:
            self._checkpoints = OrderedDict()

        self._on_reset(trainer=self, training=True)

    def fit(self,
            x_train: Union[Any, Tuple[Any]] = None,
            y_train: Any = None,
            x_val: Union[Any, Tuple[Any]] = None,
            y_val: Any = None,
            *,
            training_dataset: DataLoader = None,
            validation_dataset: DataLoader = None,
            epochs: int = None,
            checkpoint: Union[bool, int, Callable[[int], Tuple[bool, str]]] = None,
            **model_params):
        """
        Train the Neural Network model

        Parameters
        ----------
        x_train: Union[torch.Tensor, Tuple[torch.Tensor]]
            Training data. Will be ignored will``training_dataset`` is given.
        y_train: torch.Tensor
            Test data. Will be ignored will``training_dataset`` is given.
        training_dataset: DataLoader
            Torch DataLoader. If given, will only use this as training dataset.
            When loop over this dataset, it should yield a tuple contains ``x_train`` and ``y_train`` in order.
        x_val : Union[Any, Tuple[Any]]
            Data for validation.
        y_val : Any
            Data for validation.
        validation_dataset: DataLoader
        epochs : int
            Epochs. If not ``None``, it will overwrite ``self.epochs`` temporarily.
        checkpoint: Union[bool, int, Callable[[int], bool]]
            If ``True``, will save model states at each step.
            If ``int``, will save model states every `checkpoint` steps.
            If ``Callable``, the function should take current ``total_epochs`` as input return ``bool``.
        model_params: dict
            Other model parameters.
        """
        if epochs is None:
            epochs = self.epochs

        prob = self._total_epochs
        with tqdm(total=epochs, desc='Training') as pbar:
            for _ in self(x_train=x_train, y_train=y_train, x_val=x_val, y_val=y_val, training_dataset=training_dataset,
                          validation_dataset=validation_dataset, epochs=epochs, checkpoint=checkpoint,
                          **model_params):
                delta = self._total_epochs - prob
                if delta:
                    prob = self._total_epochs
                    pbar.update(delta)

    def __call__(self,
                 x_train: Union[torch.Tensor, Tuple[torch.Tensor]] = None,
                 y_train: Any = None,
                 x_val: Union[Any, Tuple[Any]] = None,
                 y_val: Any = None,
                 *,
                 training_dataset: DataLoader = None,
                 validation_dataset: DataLoader = None,
                 epochs: int = None,
                 checkpoint: Union[bool, int, Callable[[int], bool]] = None,
                 **model_params):
        """
        Train the Neural Network model

        Parameters
        ----------
        x_train
            Training data. Will be ignored will ``training_dataset`` is given.
        y_train
            Test data. Will be ignored will ``training_dataset`` is given.
        training_dataset: DataLoader
            Torch DataLoader. If given, will only use this as training dataset.
            When loop over this dataset, it should yield a tuple contains ``x_train`` and ``y_train`` in order.
        x_val : Union[Any, Tuple[Any]]
            Data for validation.
        y_val : Any
            Data for validation.
        validation_dataset : DataLoader
        epochs : int
            Epochs. If not ``None``, it will overwrite ``self.epochs`` temporarily.
        checkpoint: Union[bool, int, Callable[[int], bool]]
            If ``True``, will save model states at each step.
            If ``int``, will save model states every `checkpoint` steps.
            If ``Callable``, the function should take current ``total_epochs`` as input return ``bool``.
        model_params: dict
            Other model parameters.

        Yields
        ------
        namedtuple

        """
        if self._model is None:
            raise RuntimeError(
                'no model for training, use `trainer.model = <model>` or `trainer.reset(to=<model>)` to set one')
        if self._loss_func is None:
            raise RuntimeError(
                'no loss function for training, use `trainer.loss_func = <loss_func>` to set one')
        if self._optimizer is None:
            raise RuntimeError(
                'no optimizer for training, use `trainer.optimizer = <optimizer>` to set one')

        if epochs is None:
            epochs = self.epochs

        if training_dataset is not None:
            if y_train is not None or x_train is not None:
                raise RuntimeError('parameter <training_dataset> is exclusive of <x_train> and <y_train>')
        else:
            if y_train is None or x_train is None:
                raise RuntimeError('missing parameter <x_train> or <y_train>')

        # training step
        def _step(x_, y_, i_b=0):
            def closure():
                self._optimizer.zero_grad()
                y_p_ = self._model(*x_, **model_params)
                y_p_, y_t_ = self.output_proc(y_p_, y_, trainer=self, training=True)
                loss_ = self._loss_func(y_p_, y_t_)
                loss_.backward()

                if self._clip_grad is not None:
                    self._clip_grad(self._model.parameters())

                return loss_

            # make sure running in training mode
            if not self._model.training:
                self._model.train()

            train_loss = self._optimizer.step(closure).item()

            step_info = OrderedDict(
                total_iters=self._total_its,
                i_epoch=self._total_epochs,
                i_batch=i_b + 1, )
            step_info[self._loss_type] = train_loss
            self._total_its += 1
            self._step_forward(step_info=step_info, trainer=self, training=True)
            self._training_info.append(step_info)

            if self._lr_scheduler is not None:
                if isinstance(self._lr_scheduler, ReduceLROnPlateau):
                    self._lr_scheduler.step(train_loss, epoch=self._total_epochs)
                else:
                    self._lr_scheduler.step(epoch=self._total_its)

            return step_info

        def _snapshot():
            if checkpoint is not None:
                if isinstance(checkpoint, bool) and checkpoint:
                    self.set_checkpoint()
                if isinstance(checkpoint, int):
                    if self._total_epochs % checkpoint == 0:
                        self.set_checkpoint()
                if callable(checkpoint):
                    flag, msg = checkpoint(self._total_epochs)
                    if flag:
                        self.set_checkpoint(msg)

        if validation_dataset is not None:
            if y_val is not None or x_val is not None:
                raise RuntimeError('parameter <validation_dataset> is exclusive of <x_val> and <y_val>')
            else:
                self._validate_dataset = validation_dataset
        else:
            if y_val is not None and x_val is not None:
                self._x_val, self._y_val = self.input_proc(x_val, y_val, trainer=self, training=False)

        # before processing
        self._before_proc(trainer=self, training=True)

        if training_dataset:
            for i_epoch in range(self._total_epochs, epochs + self._total_epochs):

                self._total_epochs += 1
                for i_batch, (x_train, y_train) in enumerate(training_dataset):
                    x_train, y_train = self.input_proc(x_train, y_train, trainer=self, training=True)
                    if not isinstance(x_train, tuple):
                        x_train = (x_train,)
                    yield _step(x_train, y_train, i_batch)
                    if self._early_stopping[0]:
                        tqdm.write(f'Early stopping is applied: {self._early_stopping[1]}')
                        self._after_proc(trainer=self, training=True)
                        self._model.eval()
                        return
                _snapshot()

        else:
            x_train, y_train = self.input_proc(x_train, y_train, trainer=self, training=True)
            if not isinstance(x_train, tuple):
                x_train = (x_train,)

            for i_epoch in range(self._total_epochs, epochs + self._total_epochs):

                self._total_epochs += 1
                yield _step(x_train, y_train)
                if self._early_stopping[0]:
                    tqdm.write(f'Early stopping is applied: {self._early_stopping[1]}.')
                    self._after_proc(trainer=self, training=True)
                    self._model.eval()
                    return
                _snapshot()

        # after processing
        self._after_proc(trainer=self, training=True)
        self._model.eval()

    @classmethod
    def load(cls, from_: Union[str, Path, Checker], *,
             loss_func: torch.nn.Module = None,
             optimizer: BaseOptimizer = None,
             lr_scheduler: BaseLRScheduler = None,
             clip_grad: Union[ClipNorm, ClipValue] = None,
             epochs: int = 200,
             cuda: Union[bool, str, torch.device] = False,
             non_blocking: bool = False,
             ) -> 'Trainer':
        """
        Load model for local path or :class:`xenonpy.model.training.Checker`.

        Parameters
        ----------
        from_
            Path to the model dir or :class:`xenonpy.model.training.Checker` object.
        loss_func
            Loss function.
        optimizer
            Optimizer for model parameters tuning.
        lr_scheduler
            Learning rate scheduler.
        clip_grad
            Clip grad before each optimize.
        epochs
            Number of iterations.
        cuda
            Set training device(s).
        non_blocking
            When non_blocking is ``True``,
            it tries to convert/move asynchronously with respect to the host if possible.

        Returns
        -------

        """
        if isinstance(from_, (str, Path)):
            checker = Checker(from_)
        else:
            checker = from_
        if len(checker.files) == 0:
            raise RuntimeError(f'{checker.path} is not a model dir')

        tmp = cls(model=checker.model, cuda=cuda, loss_func=loss_func, optimizer=optimizer, lr_scheduler=lr_scheduler,
                  clip_grad=clip_grad, epochs=epochs, non_blocking=non_blocking)
        tmp._training_info = checker.training_info.to_dict(orient='records')
        if Path(checker.path + '/checkpoints').is_dir():
            for k in checker.checkpoints.files:
                tmp._checkpoints[k] = cls.checkpoint_tuple(**checker.checkpoints[k])
        return tmp

    def predict(self,
                x_in: Union[Any, Tuple[Any]] = None,
                y_true: Union[Any, Tuple[Any]] = None,
                *,
                dataset: DataLoader = None,
                checkpoint: Union[int, str] = None,
                **model_params):
        """
        Predict from x input.
        This is just a simple wrapper for :meth:`~model.nn.utils.Predictor.predict`.

        Parameters
        ----------
        checkpoint: Union[int, str]
        x_in: Union[Any, Tuple[Any]]
            Input data for prediction.
        y_true: Union[Any, Tuple[Any]]
        dataset: DataLoader
        model_params: dict
            Model parameters for prediction.

        Returns
        -------
        ret: T_Prediction
            Predict results.
        """

        def _predict(x_, y_=None):
            x_, y_ = self.input_proc(x_, y_, trainer=self, training=False)
            if not isinstance(x_, tuple):
                x_ = (x_,)

            if checkpoint:
                cp = self.get_checkpoint(checkpoint)
                model = deepcopy(self._model).to(self.device)
                model.load_state_dict(cp.model_state)
                y_p_ = model(*x_, **model_params)
            else:
                y_p_ = self._model(*x_, **model_params)

            return self.output_proc(y_p_, y_, trainer=self, training=False)

        def _vstack(ls):
            if isinstance(ls[0], np.ndarray):
                return np.vstack(ls)
            if isinstance(ls[0], torch.Tensor):
                return torch.cat(ls, dim=0)
            return ls

        # maker sure eval mode
        self._model.eval()

        if x_in is None and y_true is None and dataset is not None:
            y_preds = []
            y_trues = []
            for x_in, y_true in dataset:
                y_pred, y_true = _predict(x_in, y_true)
                y_preds.append(y_pred)
                y_trues.append(y_true)
            return _vstack(y_preds), _vstack(y_trues)
        elif x_in is not None and dataset is None:
            y_preds, y_trues = _predict(x_in, y_true)
            if y_trues is None:
                return y_preds
            return y_preds, y_trues
        else:
            raise RuntimeError('parameters <x_in> and <dataset> are mutually exclusive')

    def to_namedtuple(self):
        return self.results_tuple(
            total_epochs=self.total_epochs,
            device=self.device,
            training_info=self.training_info,
            checkpoints={k: deepcopy(v.model_state) for k, v in self._checkpoints.items()},
            model=deepcopy(self._model.cpu())
        )



#  Copyright (c) 2019. stewu5. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import pandas as pd
from rdkit import Chem
from xenonpy.inverse.iqspr import NGram

def combine_fragments(smis_base, smis_frag):
    """
    combine two SMILES strings with '*' as connection points, note that no proper treatment for '-*', '=*', or '#*' yet.

    Parameters
    ----------
    smis_base: str
        SMILES for combining.
        If no '*', assume connection point at the end.
        If more than one '*', the first will be picked if it's not the 1st character.
    smis_frag: str
        SMILES for combining.
        If no '*', assume connection point at the front.
        If more than one '*', the first will be picked.
    """

    # prepare NGram object for use of ext. SMILES
    ngram = NGram()

    # check position of '*'
    mols_base = Chem.MolFromSmiles(smis_base)
    if mols_base is None:
        raise RuntimeError('Invalid base SMILES!')
    idx_base = [i for i in range(mols_base.GetNumAtoms()) if mols_base.GetAtomWithIdx(i).GetSymbol() == '*']

    # rearrange base SMILES to avoid 1st char = '*' (assume no '**')
    if len(idx_base) == 1 and idx_base[0] == 0:
        smis_base_head = Chem.MolToSmiles(mols_base,rootedAtAtom=1)
    elif len(idx_base) == 0:
        smis_base_head = smis_base + '*'
    else:
        smis_base_head = smis_base

    # converge base to ext. SMILES and pick insertion location
    esmi_base = ngram.smi2esmi(smis_base_head)
    esmi_base = esmi_base[:-1]
    idx_base = esmi_base.index[esmi_base['esmi'] == '*'].tolist()
    if idx_base[0] == 0:
        if len(idx_base) == 1:
            # put treatment here
            raise RuntimeError('Probably -*, =*, and/or #* exist')
        else:
            idx_base = idx_base[1]
    else:
        idx_base = idx_base[0]

    # rearrange fragment to have 1st char = '*' and convert to ext. SMILES
    mols_frag = Chem.MolFromSmiles(smis_frag)
    if mols_frag is None:
        raise RuntimeError('Invalid frag SMILES!')
    idx_frag = [i for i in range(mols_frag.GetNumAtoms()) if mols_frag.GetAtomWithIdx(i).GetSymbol() == '*']
    if len(idx_frag) == 0: # if -*, =*, and/or #* exist, not counted as * right now
        esmi_frag = ngram.smi2esmi(smis_frag)
        # remove last '!'
        esmi_frag = esmi_frag[:-1]
    else:
        esmi_frag = ngram.smi2esmi(Chem.MolToSmiles(mols_frag,rootedAtAtom=idx_frag[0]))
        # remove leading '*' and last '!'
        esmi_frag = esmi_frag[1:-1]

    # check open rings of base SMILES
    nRing_base = esmi_base['n_ring'].loc[idx_base]

    # re-number rings in fragment SMILES
    esmi_frag['n_ring'] = esmi_frag['n_ring'] + nRing_base

    # delete '*' at the insertion location
    esmi_base = esmi_base.drop(idx_base).reset_index(drop=True)

    # combine base with the fragment
    ext_smi = pd.concat([esmi_base.iloc[:idx_base], esmi_frag, esmi_base.iloc[idx_base:]]).reset_index(drop=True)
    new_pd_row = {'esmi': '!', 'n_br': 0, 'n_ring': 0, 'substr': ['!']}
    ext_smi.append(new_pd_row, ignore_index=True)

    return ngram.esmi2smi(ext_smi)





#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import re
import warnings
from copy import deepcopy

import numpy as np
import pandas as pd
from rdkit import Chem
from tqdm import tqdm

from xenonpy.inverse.base import BaseProposal, ProposalError


class GetProbError(ProposalError):

    def __init__(self, tmp_str, i_b, i_r):
        self.tmp_str = tmp_str
        self.iB = i_b
        self.iR = i_r
        self.old_smi = None

        super().__init__('get_prob: %s not found in NGram, iB=%i, iR=%i' % (tmp_str, i_b, i_r))


class MolConvertError(ProposalError):
    def __init__(self, new_smi):
        self.new_smi = new_smi
        self.old_smi = None

        super().__init__('can not convert %s to Mol' % new_smi)


class NGramTrainingError(ProposalError):
    def __init__(self, error, smi):
        self.old_smi = smi

        super().__init__('training failed for %s, because of <%s>: %s' % (smi, error.__class__.__name__, error))


class NGram(BaseProposal):
    def __init__(self, *, ngram_tab=None, sample_order=(1, 10), del_range=(1, 10), min_len = 1, max_len=1000, reorder_prob=0):
        """
        N-Garm

        Parameters
        ----------
        ngram_tab: NGram table
            NGram table for modify SMILES.
        sample_order: tuple[int, int] or int
            range of order of ngram table used during proposal,
            when given int, sample_order = (1, int)
        del_range: tuple[int, int] or int
            range of random deletion of SMILES string during proposal,
            when given int, del_range = (1, int)
        min_len: int
            minimum length of the extended SMILES,
            shall be smaller than the lower bound of the sample_order
        max_len: int
            max length of the extended SMILES to be terminated from continuing modification
        reorder_prob: float
            probability of the SMILES being reordered during proposal
        """

        self.sample_order = sample_order
        self.reorder_prob = reorder_prob
        self.min_len = min_len
        self.max_len = max_len
        self.del_range = del_range

        if ngram_tab is None:
            self._table = None
            self._train_order = None
        else:
            self._table = deepcopy(ngram_tab)
            self._train_order = (1, len(ngram_tab))

        self._fit_sample_order()
        self._fit_min_len()

    @property
    def sample_order(self):
        return self._sample_order

    @sample_order.setter
    def sample_order(self, val):
        if isinstance(val, int):
            self._sample_order = (1, val)
        elif isinstance(val, tuple):
            self._sample_order = val
        elif isinstance(val, (list, np.array, pd.Series)):
            self._sample_order = (val[0], val[1])
        else:
            raise TypeError('please input a <tuple> of two <int> or a single <int> for sample_order')
        if self._sample_order[0] < 1:
            raise RuntimeError('Min sample_order must be greater than 0')
        if self._sample_order[1] < self._sample_order[0]:
            raise RuntimeError('Min sample_order must not be smaller than max sample_order')

    @property
    def reorder_prob(self):
        return self._reorder_prob

    @reorder_prob.setter
    def reorder_prob(self, val):
        if isinstance(val, (int, float)):
            self._reorder_prob = val
        else:
            raise TypeError('please input a <float> for reorder_prob')

    @property
    def min_len(self):
        return self._min_len

    @min_len.setter
    def min_len(self, val):
        if isinstance(val, int):
            self._min_len = val
        else:
            raise TypeError('please input a <int> for min_len')

    @property
    def max_len(self):
        return self._max_len

    @max_len.setter
    def max_len(self, val):
        if isinstance(val, int):
            self._max_len = val
        else:
            raise TypeError('please input a <int> for max_len')

    @property
    def del_range(self):
        return self._del_range

    @del_range.setter
    def del_range(self, val):
        if isinstance(val, int):
            self._del_range = (1, val)
        elif isinstance(val, tuple):
            self._del_range = val
        elif isinstance(val, (list, np.array, pd.Series)):
            self._del_range = (val[0], val[1])
        else:
            raise TypeError('please input a <tuple> of two <int> or a single <int> for del_range')
        if self._del_range[1] < self._del_range[0]:
            raise RuntimeError('Min del_range must not be smaller than max del_range')

    def _fit_sample_order(self):
        if self._train_order and self._train_order[1] < self.sample_order[1]:
            warnings.warn('max <sample_order>: %s is greater than max <train_order>: %s,'
                          'max <sample_order> will be reduced to max <train_order>' % (self.sample_order[1], self._train_order[1]),
                          RuntimeWarning)
            self.sample_order = (self.sample_order[0], self._train_order[1])
        if self._train_order and self._train_order[0] > self.sample_order[0]:
            warnings.warn('min <sample_order>: %s is smaller than min <train_order>: %s,'
                          'min <sample_order> will be increased to min <train_order>' % (self.sample_order[0], self._train_order[0]),
                          RuntimeWarning)
            self.sample_order = (self._train_order[0], self.sample_order[1])

    def _fit_min_len(self):
        if self.sample_order[0] > self.min_len:
            warnings.warn('min <sample_order>: %s is greater than min_len: %s,'
                          'min_len will be increased to min <sample_order>' % (
                          self.sample_order[0], self.min_len),
                          RuntimeWarning)
            self.min_len = self.sample_order[0]

    def on_errors(self, error):
        """

        Parameters
        ----------
        error: ProposalError
            Error object.
        Returns
        -------

        """
        if isinstance(error, GetProbError):
            return error.old_smi
        if isinstance(error, MolConvertError):
            return error.old_smi
        if isinstance(error, NGramTrainingError):
            pass

    @property
    def ngram_table(self):
        return deepcopy(self._table)

    @ngram_table.setter
    def ngram_table(self, value):
        self._table = deepcopy(value)

    def modify(self, ext_smi):
        # reorder for a given probability
        if np.random.random() < self.reorder_prob:
            ext_smi = self.reorder_esmi(ext_smi)
        # number of deletion (randomly pick from given range)
        n_del = np.random.randint(self.del_range[0], self.del_range[1] + 1)
        # first delete then add
        ext_smi = self.del_char(ext_smi, min(n_del + 1, len(ext_smi) - self.min_len))  # at least leave min_len char
        # add until reaching '!' or a given max value
        for i in range(self.max_len - len(ext_smi)):
            ext_smi, _ = self.sample_next_char(ext_smi)
            if ext_smi['esmi'].iloc[-1] == '!':
                return ext_smi  # stop when hitting '!', assume must be valid SMILES
        # check incomplete esmi
        ext_smi = self.validator(ext_smi)
        # fill in the '!'
        new_pd_row = {'esmi': '!', 'n_br': 0, 'n_ring': 0, 'substr': ext_smi['substr'].iloc[-1] + ['!']}

        warnings.warn('Extended SMILES hits max length', RuntimeWarning)

        return ext_smi.append(new_pd_row, ignore_index=True)

    @classmethod
    def smi2list(cls, smiles):
        # smi_pat = r'(-\[.*?\]|=\[.*?\]|#\[.*?\]|\[.*?\]|-Br|=Br|#Br|-Cl|=Cl|#Cl|Br|Cl|-.|=.|#.|\%[0-9][0-9]|\w|\W)'
        # smi_pat = r'(=\[.*?\]|#\[.*?\]|\[.*?\]|=Br|#Br|=Cl|#Cl|Br|Cl|=.|#.|\%[0-9][0-9]|\w|\W)'
        # smi_pat = r'(\[.*?\]|Br|Cl|\%[0-9][0-9]|\w|\W)'
        smi_pat = r'(\[.*?\]|Br|Cl|(?<=%)[0-9][0-9]|\w|\W)'

        # smi_list = list(filter(None, re.split(smi_pat, smiles)))
        smi_list = list(filter(lambda x: not ((x == "") or (x == "%")), re.split(smi_pat, smiles)))

        # combine bond with next token only if the next token isn't a number
        # assume SMILES does not end with a bonding character!
        tmp_idx = [i for i, x in enumerate(smi_list) if ((x in "-=#") and (not smi_list[i + 1].isdigit()))]
        if len(tmp_idx) > 0:
            for i in tmp_idx:
                smi_list[i + 1] = smi_list[i] + smi_list[i + 1]
            smi_list = np.delete(smi_list, tmp_idx).tolist()

        return smi_list

    @classmethod
    def smi2esmi(cls, smi):
        smi_list = cls.smi2list(smi)

        esmi_list = smi_list + ['!']
        substr_list = []  # list of all contracted substrings (include current char.)

        # list of whether open branch exist at current character position (include current char.)
        br_list = []

        # list of number of open ring at current character position (include current char.)
        ring_list = []
        v_substr = []  # list of temporary contracted substrings
        v_ringn = []  # list of numbering of open rings
        c_br = 0  # tracking open branch steps for recording contracted substrings
        n_br = 0  # tracking number of open branches
        tmp_ss = []  # list of current contracted substring
        for i in range(len(esmi_list)):
            if c_br == 2:
                v_substr.append(deepcopy(tmp_ss))  # contracted substring added w/o ')'
                c_br = 0
            elif c_br == 1:
                c_br = 2

            if esmi_list[i] == '(':
                c_br = 1
                n_br += 1
            elif esmi_list[i] == ')':
                tmp_ss = deepcopy(v_substr[-1])  # retrieve contracted substring added w/o ')'
                v_substr.pop()
                n_br -= 1
            elif esmi_list[i].isdigit():
                esmi_list[i] = int(esmi_list[i])
                if esmi_list[i] in v_ringn:
                    esmi_list[i] = v_ringn.index(esmi_list[i])
                    v_ringn.pop(esmi_list[i])
                else:
                    v_ringn.insert(0, esmi_list[i])
                    esmi_list[i] = '&'

            tmp_ss.append(esmi_list[i])
            substr_list.append(deepcopy(tmp_ss))
            br_list.append(n_br)
            ring_list.append(len(v_ringn))

        return pd.DataFrame({
            'esmi': esmi_list,
            'n_br': br_list,
            'n_ring': ring_list,
            'substr': substr_list
        })

    # may add error check here in the future?
    @classmethod
    def esmi2smi(cls, ext_smi):
        smi_list = ext_smi['esmi'].tolist()
        num_open = []
        num_unused = list(range(99, 0, -1))
        for i in range(len(smi_list)):
            if smi_list[i] == '&':
                if num_unused[-1] > 9:
                    smi_list[i] = ''.join(['%', str(num_unused[-1])])
                else:
                    smi_list[i] = str(num_unused[-1])
                num_open.insert(0, num_unused[-1])
                num_unused.pop()
            elif isinstance(smi_list[i], int):
                tmp = int(smi_list[i])
                if num_open[tmp] > 9:
                    smi_list[i] = ''.join(['%', str(num_open[tmp])])
                else:
                    smi_list[i] = str(num_open[tmp])
                num_unused.append(num_open[tmp])
                num_open.pop(tmp)
        if smi_list[-1] == "!":  # cover cases of incomplete esmi_pd
            smi_list.pop()  # remove the final '!'
        return ''.join(smi_list)

    def remove_table(self, max_order = None):
        """
        Remove estimators from estimator set.

        Parameters
        ----------
        max_order: int
            max order to be left in the table, the rest is removed.
        """
        if max_order:
            tmp = self._train_order[1] - max_order
            if tmp < 1:
                warnings.warn('Nothing removed', RuntimeWarning)
            else:
                self._table = self._table[:-tmp]
                self._train_order = (self._train_order[0], max_order)
        else:
            self._table = None
            self._train_order = None

    def fit(self, smiles, *, train_order=(1, 10)):
        """

        Parameters
        ----------
        smiles: list[str]
            SMILES for training.
        train_order: tuple[int, int] or int
            range of order when train a NGram table,
            when given int, train_order = (1, int),
            and train_order[0] must be > 0

        Returns
        -------

        """

        def _fit_one(ext_smi):
            for iB in [0, 1]:
                # index for open/closed branches char. position, remove last row for '!'
                idx_B = ext_smi.iloc[:-1].index[(ext_smi['n_br'].iloc[:-1] > 0) == iB]
                list_R = ext_smi['n_ring'][idx_B].unique().tolist()
                if len(list_R) > 0:
                    # expand list of dataframe for max. num-of-ring + 1
                    if len(self._table[0][iB]) < (max(list_R) + 1):
                        for ii in range(len(self._table)):
                            self._table[ii][iB].extend([
                                pd.DataFrame()
                                for i in range((max(list_R) + 1) - len(self._table[ii][iB]))
                            ])
                    for iR in list_R:
                        # index for num-of-open-ring char. pos.
                        idx_R = idx_B[ext_smi['n_ring'][idx_B] == iR]

                        # shift one down for 'next character given substring'
                        tar_char = ext_smi['esmi'][idx_R + 1].tolist()
                        tar_substr = ext_smi['substr'][idx_R].tolist()

                        for iO in range(self._train_order[0]-1, self._train_order[1]):
                            # index for char with substring length not less than order
                            idx_O = [x for x in range(len(tar_substr)) if len(tar_substr[x]) > iO]
                            for iC in idx_O:
                                if not tar_char[iC] in self._table[iO][iB][iR].columns.tolist():
                                    self._table[iO][iB][iR][tar_char[iC]] = 0
                                tmp_row = str(tar_substr[iC][-(iO + 1):])
                                if tmp_row not in self._table[iO][iB][iR].index.tolist():
                                    self._table[iO][iB][iR].loc[tmp_row] = 0

                                # somehow 'at' not ok with mixed char and int column names
                                self._table[iO][iB][iR].loc[tmp_row, tar_char[iC]] += 1

        if self._table:
            raise RuntimeError('NGram table existed.'
                               'If you want to re-train the table,'
                               'please use `remove_table()` method first.')

        if isinstance(train_order, int):
            tmp_train_order = (1, train_order)
        elif isinstance(train_order, tuple):
            tmp_train_order = train_order
        elif isinstance(train_order, (list,np.array,pd.Series)):
            tmp_train_order = (train_order[0],train_order[1])
        else:
            raise TypeError('please input a <tuple> of two <int> or a single <int> for train_order')

        if tmp_train_order[0] < 1:
            raise RuntimeError('Min train_order must be greater than 0')
        if tmp_train_order[1] < tmp_train_order[0]:
            raise RuntimeError('Min train_order must not be smaller than max train_order')

        self._train_order = tmp_train_order
        self._table = [[[], []] for _ in range(self._train_order[1])]

        self._fit_sample_order()
        self._fit_min_len()
        for smi in tqdm(smiles):
            try:
                _fit_one(self.smi2esmi(smi))
            except Exception as e:
                warnings.warn('NGram training failed for %s' % smi, RuntimeWarning)
                e = NGramTrainingError(e, smi)
                self.on_errors(e)

        return self

    # get probability vector for sampling next character, return character list and corresponding probability in numpy.array (normalized)
    # may cause error if empty string list is fed into 'tmp_str'
    # Warning: maybe can reduce the input of iB and iR - directly input the reduced list of self._ngram_tab (?)
    # Warning: may need to update this function with bisection search for faster speed (?)
    # Warning: may need to add worst case that no pattern found at all?

    def get_prob(self, tmp_str, iB, iR):
        # right now we use back-off method, an alternative is Kneser–Nay smoothing
        cand_char = []
        cand_prob = 1
        iB = int(iB)
        for iO in range(self.sample_order[1] - 1, self.sample_order[0] - 2, -1):
            # if (len(tmp_str) > iO) & (str(tmp_str[-(iO + 1):]) in self._table[iO][iB][iR].index.tolist()):
            if len(tmp_str) > iO and str(tmp_str[-(iO + 1):]) in self._table[iO][iB][iR].index.tolist():
                cand_char = self._table[iO][iB][iR].columns.tolist()
                cand_prob = np.array(self._table[iO][iB][iR].loc[str(tmp_str[-(iO + 1):])])
                break
        if len(cand_char) == 0:
            warnings.warn('get_prob: %s not found in NGram, iB=%i, iR=%i' % (tmp_str, iB, iR), RuntimeWarning)
            raise GetProbError(tmp_str, iB, iR)
        return cand_char, cand_prob / np.sum(cand_prob)

    # get the next character, return the probability value
    def sample_next_char(self, ext_smi):
        iB = ext_smi['n_br'].iloc[-1] > 0
        iR = ext_smi['n_ring'].iloc[-1]
        cand_char, cand_prob = self.get_prob(ext_smi['substr'].iloc[-1], iB, iR)
        # here we assume cand_char is not empty
        idx = np.random.choice(range(len(cand_char)), p=cand_prob)
        next_char = cand_char[idx]
        ext_smi = self.add_char(ext_smi, next_char)
        return ext_smi, cand_prob[idx]

    @classmethod
    def add_char(cls, ext_smi, next_char):
        new_pd_row = ext_smi.iloc[-1]
        new_pd_row.at['substr'] = new_pd_row['substr'] + [next_char]
        new_pd_row.at['esmi'] = next_char
        if next_char == '(':
            new_pd_row.at['n_br'] += 1
        elif next_char == ')':
            new_pd_row.at['n_br'] -= 1
            # assume '(' must exist before if the extended SMILES is valid! (will fail if violated)
            # idx = next((x for x in range(len(new_pd_row['substr'])-1,-1,-1) if new_pd_row['substr'][x] == '('), None)
            # find index of the last unclosed '('
            tmp_c = 1
            for x in range(len(new_pd_row['substr']) - 2, -1, -1):  # exclude the already added "next_char"
                if new_pd_row['substr'][x] == '(':
                    tmp_c -= 1
                elif new_pd_row['substr'][x] == ')':
                    tmp_c += 1
                if tmp_c == 0:
                    idx = x
                    break
            # assume no '()' and '((' pattern that is not valid/possible in SMILES
            new_pd_row.at['substr'] = new_pd_row['substr'][:(idx + 2)] + [')']
        elif next_char == '&':
            new_pd_row.at['n_ring'] += 1
        elif isinstance(next_char, int):
            new_pd_row.at['n_ring'] -= 1
        return ext_smi.append(new_pd_row, ignore_index=True)

    @classmethod
    def del_char(cls, ext_smi, n_char):
        if n_char > 0:
            return ext_smi[:-n_char]
        else:
            return ext_smi

    # need to make sure esmi_pd is a completed SMILES to use this function
    # todo: kekuleSmiles?
    @classmethod
    def reorder_esmi(cls, ext_smi):
        # convert back to SMILES first, then to rdkit MOL
        mol = Chem.MolFromSmiles(cls.esmi2smi(ext_smi))
        idx = np.random.choice(range(mol.GetNumAtoms())).item()
        # currently assume kekuleSmiles=True, i.e., no small letters but with ':' for aromatic rings
        ext_smi = cls.smi2esmi(Chem.MolToSmiles(mol, rootedAtAtom=idx))

        return ext_smi

    def validator(self, ext_smi):
        # delete all ending '(' or '&'
        for i in range(len(ext_smi)):
            if not ((ext_smi['esmi'].iloc[-1] == '(') or (ext_smi['esmi'].iloc[-1] == '&')):
                break
            ext_smi = self.del_char(ext_smi, 1)
        # delete or fill in ring closing
        flag_ring = ext_smi['n_ring'].iloc[-1] > 0
        for i in range(len(ext_smi)):  # max to double the length of current SMILES
            if flag_ring and (np.random.random() < 0.7):  # 50/50 for adding two new char.
                # add a character
                ext_smi, _ = self.sample_next_char(ext_smi)
                flag_ring = ext_smi['n_ring'].iloc[-1] > 0
            else:
                break
        if flag_ring:
            # prepare for delete (1st letter shall not be '&')
            tmp_idx = ext_smi.iloc[1:].index
            tmp_count = np.array(ext_smi['n_ring'].iloc[1:]) - np.array(ext_smi['n_ring'].iloc[:-1])
            num_open = tmp_idx[tmp_count == 1].values.tolist()
            num_open.reverse()
            num_close = tmp_idx[tmp_count == -1].values.tolist()
            idx_pop = []
            for i in num_close:
                idx_pop.append(ext_smi['esmi'][i])
            for ii, i in enumerate(idx_pop):
                ext_smi['esmi'][num_close[ii]] += sum([x < i for x in idx_pop[ii+1:]]) - i 
                num_open.pop(i)
            # delete all irrelevant rows and reconstruct esmi
            ext_smi = self.smi2esmi(
                self.esmi2smi(ext_smi.drop(ext_smi.index[num_open]).reset_index(drop=True)))
            ext_smi = ext_smi.iloc[:-1]  # remove the '!'

            # delete ':' that are not inside a ring
        # tmp_idx = esmi_pd.index[(esmi_pd['esmi'] == ':') & (esmi_pd['n_ring'] < 1)]
        # if len(tmp_idx) > 0:
        #     esmi_pd = smi2esmi(esmi2smi(esmi_pd.drop(tmp_idx).reset_index(drop=True)))
        #     esmi_pd = esmi_pd.iloc[:-1] # remove the '!'
        # fill in branch closing (last letter shall not be '(')
        for i in range(ext_smi['n_br'].iloc[-1]):
            ext_smi = self.add_char(ext_smi, ')')

        return ext_smi

    def proposal(self, smiles):
        """
        Propose new SMILES based on the given SMILES.

        Parameters
        ----------
        smiles: list of SMILES
            Given SMILES for modification.

        Returns
        -------
        new_smiles: list of SMILES
            The proposed SMILES from the given SMILES.
        """
        new_smis = []
        for i, smi in enumerate(smiles):
            ext_smi = self.smi2esmi(smi)
            try:
                new_ext_smi = self.modify(ext_smi)
                new_smi = self.esmi2smi(new_ext_smi)
                if Chem.MolFromSmiles(new_smi) is None:
                    warnings.warn('can not convert %s to Mol' % new_smi, RuntimeWarning)
                    raise MolConvertError(new_smi)
                new_smis.append(new_smi)

            except ProposalError as e:
                e.old_smi = smi
                new_smi = self.on_errors(e)
                new_smis.append(new_smi)

            except Exception as e:
                raise e

        return new_smis

    def _merge_table(self, ngram_tab, weight=1):
        """
        Merge with a given NGram table

        Parameters
        ----------
        ngram_tab: NGram
            the table in the given NGram class variable will be merged to the table in self
        weight: double
            a scalar to scale the frequency in the given NGram table

        Returns
        -------
        tmp_n_gram: NGram
            merged NGram tables
        """

        self._train_order = (min(self._train_order[0],ngram_tab._train_order[0]), max(self._train_order[1],ngram_tab._train_order[1]))
        self._fit_sample_order()
        self._fit_min_len()

        n_gram_tab1 = self._table  # do not use deepcopy here
        n_gram_tab2 = ngram_tab.ngram_table  # default deepcopy used
        w = weight

        ord1 = len(n_gram_tab1)
        ord2 = len(n_gram_tab2)
        Bc1 = len(n_gram_tab1[0][0])
        Bc2 = len(n_gram_tab2[0][0])
        Bo1 = len(n_gram_tab1[0][1])
        Bo2 = len(n_gram_tab2[0][1])

        # fix the number of ring mis-match first
        if Bc1 < Bc2:
            for ii in range(ord1):
                n_gram_tab1[ii][0].extend([
                    pd.DataFrame()
                    for _ in range(Bc2 - Bc1)
                ])
        elif Bc1 > Bc2:
            for ii in range(ord2):
                n_gram_tab2[ii][0].extend([
                    pd.DataFrame()
                    for _ in range(Bc1 - Bc2)
                ])
        if Bo1 < Bo2:
            for ii in range(ord1):
                n_gram_tab1[ii][1].extend([
                    pd.DataFrame()
                    for _ in range(Bo2 - Bo1)
                ])
        elif Bo1 > Bo2:
            for ii in range(ord2):
                n_gram_tab2[ii][1].extend([
                    pd.DataFrame()
                    for _ in range(Bo1 - Bo2)
                ])

        # fix order mis-match
        if ord2 > ord1:
            n_gram_tab1.extend(n_gram_tab2[ord1:])

        # combine overlapped order (weighted on tab2)
        for i in range(min(ord1, ord2)):
            for j in range(len(n_gram_tab1[i])):
                for k in range(len(n_gram_tab1[i][j])):
                    n_gram_tab1[i][j][k] = n_gram_tab1[i][j][k].add(w * n_gram_tab2[i][j][k], fill_value=0).fillna(0)

    def merge_table(self, *ngram_tab: 'NGram', weight=1, overwrite=True):
        """
        Merge with a given NGram table

        Parameters
        ----------
        ngram_tab
            the table(s) in the given NGram class variable(s) will be merged to the table in self
        weight: int/float or list/tuple/np.array/pd.Series[int/float]
            a scalar/vector to scale the frequency in the given NGram table to be merged,
            must have the same length as ngram_tab
        overwrite: boolean
            overwrite the original table (self) or not,
            do not recommend to be False (may have memory issue)

        Returns
        -------
        tmp_n_gram: NGram
            merged NGram tables
        """

        if not np.all([isinstance(x, NGram) for x in ngram_tab]):
            raise TypeError('each element in the input must be <NGram>')

        if isinstance(weight, (int, float)):
            weight = np.repeat(weight, len(ngram_tab))
        elif isinstance(weight, (tuple, list, np.array, pd.Series)):
            if not np.all([isinstance(x, (int, float)) for x in weight]):
                raise TypeError('each element in weight must be <int> or <float>')
        else:
            raise TypeError('weight must be <int> or <float> or a list of them')

        if overwrite:
            tmp_n_gram = self  # do not use deepcopy here
        else:
            tmp_n_gram = deepcopy(self)

        for i, tab in enumerate(ngram_tab):
            tmp_n_gram._merge_table(ngram_tab=tab, weight=weight[i])

        return tmp_n_gram

    def split_table(self, cut_order):
        """
        Split NGram table into two

        Parameters
        ----------
        cut_order: int
            split NGram table between cut_order and cut_order+1

        Returns
        -------
        n_gram1: NGram
        n_gram2: NGram
        """

        n_gram1 = deepcopy(self)
        n_gram1.remove_table(max_order=cut_order)
        n_gram1._fit_sample_order()
        n_gram1._fit_min_len()

        n_gram2 = deepcopy(self)
        for iB in [0, 1]:
            for ii in range(cut_order):
                n_gram2._table[ii][iB] = [pd.DataFrame() for _ in range(len(n_gram2._table[ii][iB]))]
        n_gram2._train_order = (cut_order+1, self._train_order[1])
        n_gram2._fit_sample_order()
        n_gram2._fit_min_len()

        return n_gram1, n_gram2



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import json
from os import remove
from pathlib import Path
from shutil import make_archive

import pandas as pd
import requests
from requests import HTTPError
from sklearn.base import BaseEstimator

from xenonpy.utils import TimedMetaClass


class BaseQuery(BaseEstimator, metaclass=TimedMetaClass):
    queryable = None

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        if self.queryable is None:
            raise RuntimeError('Query class must give a queryable field in list of string')

        self._results = None
        self._return_json = False
        self._endpoint = endpoint
        self._api_key = api_key
        self._variables = variables

    @property
    def api_key(self):
        return self._api_key

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def variables(self):
        return self._variables

    @property
    def results(self):
        return self._results

    def gql(self, *query_vars: str):
        raise NotImplementedError()

    @staticmethod
    def _post(ret, return_json):
        if return_json:
            return ret

        if not isinstance(ret, list):
            ret = [ret]
        ret = pd.DataFrame(ret)
        return ret

    def check_query_vars(self, *query_vars: str):
        if not set(query_vars) <= set(self.queryable):
            raise RuntimeError(f'`query_vars` contains illegal variables, '
                               f'available querying variables are: {self.queryable}')
        return query_vars

    def __call__(self, *querying_vars, file=None, return_json=None):
        if len(querying_vars) == 0:
            query = self.gql(*self.queryable)
        else:
            query = self.gql(*self.check_query_vars(*querying_vars))

        payload = json.dumps({'query': query, 'variables': self._variables})

        if file is None:
            ret = requests.post(url=self._endpoint,
                                data=payload,
                                headers={"content-type": "application/json",
                                         'api_key': self._api_key})
        else:
            file = Path(file).resolve()
            file = make_archive(str(file), 'gztar', str(file))
            operations = ('operations', payload)
            maps = ('map', json.dumps({0: ['variables.model']}))
            payload_tuples = (operations, maps)
            files = {'0': open(file, 'rb')}
            try:
                ret = requests.post(url=self._endpoint,
                                    data=payload_tuples,
                                    headers={'api_key': self._api_key},
                                    files=files)
            finally:
                files['0'].close()
                remove(file)

        if ret.status_code != 200:
            try:
                message = ret.json()
            except json.JSONDecodeError:
                message = "Server did not responce."

            raise HTTPError('status_code: %s, %s' %
                            (ret.status_code, message))
        ret = ret.json()
        if 'errors' in ret:
            raise ValueError(ret['errors'][0]['message'])
        query_name = self.__class__.__name__
        ret = ret['data'][query_name[0].lower() + query_name[1:]]

        if not ret:
            return None

        if return_json is None:
            return_json = self._return_json

        ret = self._post(ret, return_json)
        self._results = ret
        return ret

    def __repr__(self, N_CHAR_MAX=700):
        queryable = '\n '.join(self.queryable)
        return f'{super().__repr__(N_CHAR_MAX)}\nQueryable: \n {queryable}'



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import warnings

import numpy as np
import pandas as pd
import torch

from xenonpy.descriptor.base import BaseFeaturizer
from xenonpy.model import SequentialLinear

__all__ = ['FrozenFeaturizer']


class FrozenFeaturizer(BaseFeaturizer):
    """
    A Featurizer to extract hidden layers a from NN model.
    """

    def __init__(self, model: torch.nn.Module = None, *,
                 cuda: bool = False, depth=None, n_layer=None, 
                 on_errors='raise', return_type='any'):
        """
        Parameters
        ----------
        model: torch.nn.Module
            Source model.
        cuda: bool
            If ``true``, run on GPU.
        depth: int
            The depth will be retrieved from NN model.
        n_layer: int
            Number of layer to be retrieved starting from the given depth.
        on_errors: string
            How to handle exceptions in feature calculations. Can be 'nan', 'keep', 'raise'.
            When 'nan', return a column with ``np.nan``.
            The length of column corresponding to the number of feature labs.
            When 'keep', return a column with exception objects.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            ``array`` and ``df`` force return type to ``np.ndarray`` and ``pd.DataFrame`` respectively.
            If ``any``, the return type dependent on the input type.
            Default is ``any``
        """
        super().__init__(n_jobs=0, on_errors=on_errors, return_type=return_type)
        self.depth = depth
        self.n_layer = n_layer
        self.model = model
        self.cuda = cuda
        self._ret = []
        self.__authors__ = ['TsumiNa']
        self.model.eval()
        self._depth = 0

    def featurize(self, descriptor, *, depth=None, n_layer=None):
        if not isinstance(self.model, torch.nn.Module):
            raise TypeError('<model> must be a instance of <torch.nn.Module>')
        hlayers = []
        if isinstance(descriptor, pd.DataFrame):
            descriptor = descriptor.values
        x_ = torch.from_numpy(descriptor).float()
        if self.cuda:
            x_.cuda()
            self.model.cuda()
        else:
            x_.cpu()
            self.model.cpu()

        if isinstance(self.model, SequentialLinear):
            for n, m in self.model.named_children():
                if 'layer_' in n:
                    hlayers.append(m.linear(x_).data)
                    x_ = m(x_)
        else:
            for m in self.model[:-1]:
                hlayers.append(m.layer(x_).data)
                x_ = m(x_)
        
        # get predefined values when depth/n_layer is not given initially
        if depth is None:
            depth = self.depth
        if n_layer is None:
            n_layer = self.n_layer

        # check updated depth values
        if depth is None: # self.depth must be None
            self.depth = len(hlayers) # update self.depth
            self._depth = len(hlayers)
        elif depth > len(hlayers):
            warnings.warn('<depth> is greater than the max depth of hidden layers')
            self._depth = len(hlayers)
        else:
            self._depth = depth

        if n_layer is None:
            l_end = 0
        else:
            l_end = n_layer-self._depth

        if l_end > -1:
            if l_end > 0:
                warnings.warn('<n_layer> is over the max depth of hidden layers starting at the given <depth>')
            ret = hlayers[-self._depth:]
        else:
            ret = hlayers[-self._depth:l_end]
        
        if self.cuda:
            ret = [l.cpu().numpy() for l in ret]
        else:
            ret = [l.numpy() for l in ret]
        self._ret = ret
        return np.concatenate(ret, axis=1)

    @property
    def feature_labels(self):
        if self._depth == 0:
            raise ValueError('Can not generate labels before transform.')
        return ['L(' + str(i - self._depth) + ')_' + str(j + 1)
                for i in range(len(self._ret))
                for j in range(self._ret[i].shape[1])]



# IQSPR with focus on molecule variety: bring in new initial molecules from reservoir in every step of SMC

import numpy as np
import pandas as pd

from xenonpy.inverse.base import BaseSMC, BaseProposal, BaseLogLikelihood

class SMCError(Exception):
    """Base exception for SMC classes"""
    pass

class IQSPR_V(BaseSMC):

    def __init__(self, *, estimator, modifier):
        """
        SMC iqspr runner.
        Parameters
        ----------
        estimator : BaseLogLikelihood or BaseLogLikelihoodSet
            Log likelihood estimator for given SMILES.
        modifier : BaseProposal
            Modify given SMILES to new ones.
        """
        self._proposal = modifier
        self._log_likelihood = estimator

    def resample(self, sims, size, p):
        return np.random.choice(sims, size=size, p=p)

    @property
    def modifier(self):
        return self._proposal

    @modifier.setter
    def modifier(self, value):
        self._proposal = value

    @property
    def estimator(self):
        return self._log_likelihood

    @estimator.setter
    def estimator(self, value):
        self._log_likelihood = value
        
    def __call__(self, reservoir, beta, size=100, *, samples=None, ratio=0.5, yield_lpf=False):
        """
        Run SMC
        Parameters
        ----------
        reservoir: list of object
            Samples to be drawn as new initial molecules in each step of SMC
        beta: list/1D-numpy of float or pd.Dataframe
            Annealing parameters for each step.
            If pd.Dataframe, column names should follow keys of mdl in BaseLogLikeihood or BaseLogLikelihoodSet
        size: int
            Sample size for each draw.
        samples: list of object
            Initial samples.
        ratio: float
            ratio of molecules to be replaced from reservoir in each step of SMC
        yield_lpf : bool
            Yield estimated log likelihood, probability and frequency of each samples. Default is ``False``.
        Yields
        -------
        samples: list of object
            New samples in each SMC iteration.
        llh: np.ndarray float
            Estimated values of log-likelihood of each samples.
            Only yield when ``yield_lpf=Ture``.
        p: np.ndarray of float
            Estimated probabilities of each samples.
            Only yield when ``yield_lpf=Ture``.
        freq: np.ndarray of float
            The number of unique samples in original samples.
            Only yield when ``yield_lpf=Ture``.
        """

        # initial samples will be randomly picked from resevoir if samples are provided
        if samples is None:
            samples = np.random.choice(reservoir,size=size).tolist()
        # refill samples if len(samples) not equals given size
        elif len(samples) < size:
            samples = samples + np.random.choice(reservoir,size=size-len(samples)).tolist()
            
        res_size = int(size*ratio)
        smc_size = size - res_size

        try:
            unique, frequency = self.unique(samples)
            ll = self.log_likelihood(unique)

            if isinstance(beta, pd.DataFrame):
                beta = beta[ll.columns].values
            else:
                # assume only one row for beta (1-D list or numpy vector)
                beta = np.transpose(np.repeat([beta], ll.shape[1], axis=0))

            w = np.dot(ll.values, beta[0]) + np.log(frequency)
            w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
            p = np.exp(w - w_sum)
            if yield_lpf:
                yield unique, ll, p, frequency
            else:
                yield unique

            # beta = np.delete(beta,0,0) #remove first "row"

        except SMCError as e:
            self.on_errors(0, samples, e)
        except Exception as e:
            raise e

        # translate between input representation and execute environment representation.
        # make sure beta is not changed (np.delete will return the deleted version, without changing original vector)
        for i, step in enumerate(np.delete(beta, 0, 0)):
            try:
                re_samples = self.resample(unique, smc_size, p)
                samples = self.proposal(re_samples + np.random.choice(reservoir,size=res_size).tolist()) 

                unique, frequency = self.unique(samples)

                # annealed likelihood in log - adjust with copy counts
                ll = self.log_likelihood(unique)
                w = np.dot(ll.values, step) + np.log(frequency)
                w_sum = np.log(np.sum(np.exp(w - np.max(w)))) + np.max(w)  # avoid underflow
                p = np.exp(w - w_sum)
                if yield_lpf:
                    yield unique, ll, p, frequency
                else:
                    yield unique

            except SMCError as e:
                self.on_errors(i + 1, samples, e)
            except Exception as e:
                raise e
                



#  Copyright (c) 2019. yoshida-lab. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

import pandas as pd

from xenonpy.mdl.base import BaseQuery


class QueryModelDetailsWith(BaseQuery):
    common = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang',

    ]

    classification = [
        'accuracy',
        'precision',
        'recall',
        'f1',
        'sensitivity',
        'prevalence',
        'specificity',
        'ppv',
        'npv',
    ]

    regression = [
        'meanAbsError',
        'maxAbsError',
        'meanSquareError',
        'rootMeanSquareError',
        'r2',
        'pValue',
        'spearmanCorr',
        'pearsonCorr',
    ]

    queryable = common + classification + regression

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        reg, cls = [], []
        if 'id' not in query_vars:
            query_vars = query_vars + ('id',)
        for var in query_vars:
            if var in self.common:
                # common.append(var)
                reg.append(var)
                cls.append(var)
            elif var in self.regression:
                reg.append(var)
            elif var in self.classification:
                cls.append(var)

        return f'''
            query (
                $modelset_has: [String!]
                $property_has: [String!]
                $descriptor_has: [String!]
                $method_has: [String!]
                $lang_has: [String!]
                $regression: Boolean
                $transferred: Boolean
                $deprecated: Boolean
                $succeed: Boolean
            ) {{
                queryModelDetailsWith(
                    modelset_has: $modelset_has
                    property_has: $property_has
                    descriptor_has: $descriptor_has
                    method_has: $method_has
                    lang_has: $lang_has
                    regression: $regression
                    transferred: $transferred
                    deprecated: $deprecated
                    succeed: $succeed
                ) {{
                    ...on Regression {{
                        {' '.join(reg)}
                    }}
                    ...on Classification {{
                        {' '.join(cls)}
                    }}
                }}
            }}
            '''


class QueryModelDetails(BaseQuery):
    common = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang',

    ]

    classification = [
        'accuracy',
        'precision',
        'recall',
        'f1',
        'sensitivity',
        'prevalence',
        'specificity',
        'ppv',
        'npv',
    ]

    regression = [
        'meanAbsError',
        'maxAbsError',
        'meanSquareError',
        'rootMeanSquareError',
        'r2',
        'pValue',
        'spearmanCorr',
        'pearsonCorr',
    ]
    queryable = common + classification + regression

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        reg, cls = [], []
        if 'id' not in query_vars:
            query_vars = query_vars + ('id',)
        for var in query_vars:
            if var in self.common:
                reg.append(var)
                cls.append(var)
            elif var in self.regression:
                reg.append(var)
            elif var in self.classification:
                cls.append(var)

        return f'''
            query ($query: [String!]!) {{
                queryModelDetails(query: $query) {{
                    ...on Regression {{
                        {' '.join(reg)}
                    }}
                    ...on Classification {{
                        {' '.join(cls)}
                    }}
                }}
            }}
            '''


class GetModelUrls(BaseQuery):
    queryable = [
        'id',
        'etag',
        'url',
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($ids: [Int!]!) {{
                getModelUrls(ids: $ids) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class GetModelUrl(BaseQuery):
    queryable = [
        'id',
        'etag',
        'url',
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($id: Int!) {{
                getModelUrl(id: $id) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class GetModelDetails(BaseQuery):
    common = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang',

    ]

    classification = [
        'accuracy',
        'precision',
        'recall',
        'f1',
        'sensitivity',
        'prevalence',
        'specificity',
        'ppv',
        'npv',
    ]

    regression = [
        'meanAbsError',
        'maxAbsError',
        'meanSquareError',
        'rootMeanSquareError',
        'r2',
        'pValue',
        'spearmanCorr',
        'pearsonCorr',
    ]
    queryable = common + classification + regression

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        reg, cls = [], []
        for var in query_vars:
            if var in self.common:
                reg.append(var)
                cls.append(var)
            elif var in self.regression:
                reg.append(var)
            elif var in self.classification:
                cls.append(var)

        return f'''
            query ($ids: [Int!]!) {{
                getModelDetails(ids: $ids) {{
                    ...on Regression {{
                        {' '.join(reg)}
                    }}
                    ...on Classification {{
                        {' '.join(cls)}
                    }}
                }}
            }}
            '''
        pass


class GetModelDetail(BaseQuery):
    common = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang',

    ]

    classification = [
        'accuracy',
        'precision',
        'recall',
        'f1',
        'sensitivity',
        'prevalence',
        'specificity',
        'ppv',
        'npv',
    ]

    regression = [
        'meanAbsError',
        'maxAbsError',
        'meanSquareError',
        'rootMeanSquareError',
        'r2',
        'pValue',
        'spearmanCorr',
        'pearsonCorr',
    ]
    queryable = common + classification + regression

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        reg, cls = [], []
        for var in query_vars:
            if var in self.common:
                reg.append(var)
                cls.append(var)
            elif var in self.regression:
                reg.append(var)
            elif var in self.classification:
                cls.append(var)

        return f'''
            query ($id: Int!) {{
                getModelDetail(id: $id) {{
                    ...on Regression {{
                        {' '.join(reg)}
                    }}
                    ...on Classification {{
                        {' '.join(cls)}
                    }}
                }}
            }}
            '''
        pass


class GetTrainingInfo(BaseQuery):
    queryable = []

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    @staticmethod
    def _post(ret, return_json):
        return pd.DataFrame(ret)

    def gql(self, *query_vars: str):
        return f'''
            query ($id: Int!) {{
                getTrainingInfo(id: $id) 
            }}
            '''


class GetTrainingEnv(BaseQuery):
    queryable = []

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)
        self._return_json = True

    def gql(self, *query_vars: str):
        return f'''
            query ($id: Int!) {{
                getTrainingEnv(id: $id) 
            }}
            '''


class GetSupplementary(BaseQuery):
    queryable = []

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)
        self._return_json = True

    def gql(self, *query_vars: str):
        return f'''
            query ($id: Int!) {{
                getSupplementary(id: $id) 
            }}
            '''


class ListModelsWithProperty(BaseQuery):
    queryable = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang'
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($name: String!) {{
                listModelsWithProperty(name: $name) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class ListModelsWithModelset(BaseQuery):
    queryable = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang'
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($name: String!) {{
                listModelsWithModelset(name: $name) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class ListModelsWithMethod(BaseQuery):
    queryable = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang'
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($name: String!) {{
                listModelsWithMethod(name: $name) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class ListModelsWithDescriptor(BaseQuery):
    queryable = [
        'id',
        'transferred',
        'succeed',
        'isRegression',
        'deprecated',
        'modelset',
        'method',
        'property',
        'descriptor',
        'lang'
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f'''
            query ($name: String!) {{
                listModelsWithDescriptor(name: $name) {{
                    {' '.join(query_vars)}
                }}
            }}
            '''


class UploadModel(BaseQuery):
    queryable = [
        'id',
        'etag',
        'path'
    ]

    def __init__(self, variables, *, api_key: str = 'anonymous.user.key',
                 endpoint: str = 'http://xenon.ism.ac.jp/api'):
        """
        Access to XenonPy.MDL library.

        Parameters
        ----------
        api_key
            Not implement yet.
        """
        super().__init__(variables=variables, api_key=api_key, endpoint=endpoint)

    def gql(self, *query_vars: str):
        return f"""
                mutation(
                    $id: Int!
                    $describe: UploadModelInput!
                    $model: Upload!
                    $training_env: Json
                    $training_info: Json
                    $supplementary: Json
                ) {{
                    uploadModel(
                        modelsetId: $id
                        model: $model
                        describe: $describe
                        training_env: $training_env
                        training_info: $training_info
                        supplementary: $supplementary
                    ) {{
                        {' '.join(query_vars)}
                    }}
                }}
                """



#  Copyright (c) 2019. TsumiNa. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from .cgcnn import *
from .extern import *
from .sequential import *
